"""
This code belongs in run_job.py, but this is split out to avoid circular dependencies
"""
from __future__ import annotations

import abc
import asyncio
import collections
import dataclasses
import enum
import itertools
import pickle
import time
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

import asyncssh
import cloudpickle
from typing_extensions import Literal


import meadowrun.ssh as ssh
from meadowrun.config import MEMORY_GB
from meadowrun.instance_allocation import allocate_jobs_to_instances, InstanceRegistrar
from meadowrun.instance_selection import ResourcesInternal
from meadowrun.meadowrun_pb2 import Job, ProcessState, PyAgentJob
from meadowrun.shared import unpickle_exception

if TYPE_CHECKING:
    from meadowrun.abstract_storage_bucket import AbstractStorageBucket
    from meadowrun.credentials import UsernamePassword
    from meadowrun.run_job_local import TaskWorkerServer, WorkerMonitor
    from types import TracebackType


_T = TypeVar("_T")
_U = TypeVar("_U")


CloudProvider = "EC2", "AzureVM"
CloudProviderType = Literal["EC2", "AzureVM"]


async def _retry(
    function: Callable[[], Awaitable[_T]],
    exception_types: Union[Type, Tuple[Type, ...]],
    max_num_attempts: int = 5,
    delay_seconds: float = 1,
    message: str = "Retrying on error",
) -> _T:
    i = 0
    while True:
        try:
            return await function()
        except exception_types as e:
            i += 1
            if i >= max_num_attempts:
                raise
            else:
                print(f"{message}: {e}")
                await asyncio.sleep(delay_seconds)


def _needs_cuda(flags_required: Union[Iterable[str], str, None]) -> bool:
    if flags_required is None:
        return False
    if isinstance(flags_required, str):
        return "nvidia" == flags_required
    return "nvidia" in flags_required


@dataclasses.dataclass(frozen=True)
class Resources:
    """
    Specifies the requirements for a job or for each task within a job

    Attributes:
        logical_cpu: Specifies logical CPU (aka vCPU) required. E.g. 2 means we require
            2 logical CPUs
        memory_gb: Specifies RAM required. E.g. 1.5 means we requires 1.5 GB of RAM
        max_eviction_rate: Specifies what eviction rate (aka interruption probability)
            we're okay with as a percent. E.g. `80` means that any instance type with an
            eviction rate less than 80% can be used. Use `0` to indicate that only
            on-demand instance are acceptable (i.e. do not use spot instances)
        gpus: Number of GPUs required. If gpu_memory is set, but this value is not set,
            this is implied to be 1
        gpu_memory: Total GPU memory (aka VRAM) required across all GPUs
        flags_required: E.g. "intel", "avx512", etc.
    """

    logical_cpu: Optional[float] = None
    memory_gb: Optional[float] = None
    max_eviction_rate: float = 80
    gpus: Optional[float] = None
    gpu_memory: Optional[float] = None
    flags: Union[Iterable[str], str, None] = None

    def uses_gpu(self) -> bool:
        return self.gpus is not None or self.gpu_memory is not None

    def needs_cuda(self) -> bool:
        return self.uses_gpu() and _needs_cuda(self.flags)

    def to_internal(self) -> Optional[ResourcesInternal]:
        if self.logical_cpu is None or self.memory_gb is None:
            return None
        else:
            return ResourcesInternal.from_cpu_and_memory(
                self.logical_cpu,
                self.memory_gb,
                self.max_eviction_rate,
                self.gpus,
                self.gpu_memory,
                self.flags,
            )


class WaitOption(enum.Enum):
    WAIT_AND_TAIL_STDOUT = 1
    WAIT_SILENTLY = 2
    DO_NOT_WAIT = 3


class Host(abc.ABC):
    """
    Host is an abstract class for specifying where to run a job. See implementations
    below.
    """

    async def set_defaults(self) -> None:
        # This function gives derived classes an opportunity to set defaults before any
        # other functions on the object are called.
        pass

    @abc.abstractmethod
    async def run_job(
        self,
        resources_required: Optional[ResourcesInternal],
        job: Job,
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        pass

    async def run_map(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> Optional[Sequence[_U]]:
        async_iterator = self.run_map_as_completed(
            function,
            args,
            resources_required_per_task,
            job_fields,
            num_concurrent_tasks,
            pickle_protocol,
            wait_for_result,
            max_num_task_attempts,
            retry_with_more_memory,
        )

        if wait_for_result == WaitOption.DO_NOT_WAIT:
            # we still need to iterate, even though no values are returned, to execute
            # the rest of the code in the iterator.
            async for _ in async_iterator:
                pass
            return None
        else:
            task_results: List[TaskResult] = []

            # TODO - this will wait forever if any tasks are missing
            async for task_result in async_iterator:
                task_results.append(task_result)

            task_results.sort(key=lambda tr: tr.task_id)

            # if tasks were None, we'd have throw already
            failed_tasks = [result for result in task_results if not result.is_success]
            if failed_tasks:
                raise RunMapTasksFailedException(
                    failed_tasks, [args[task.task_id] for task in failed_tasks]
                )

            return [result.result for result in task_results]  # type: ignore[misc]

    @abc.abstractmethod
    def run_map_as_completed(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> AsyncIterable[TaskResult[_U]]:
        pass

    @abc.abstractmethod
    async def get_storage_bucket(self) -> AbstractStorageBucket:
        pass


class SshHost(Host):
    """
    Tells run_function and related functions to connect to the remote machine over SSH.
    """

    def __init__(
        self,
        address: str,
        username: str,
        private_key: asyncssh.SSHKey,
        cloud_provider: Optional[Tuple[CloudProviderType, str]] = None,
        instance_name: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.address = address
        self.username = username
        self.private_key = private_key
        # If this field is populated, it will be a tuple of (cloud provider, region
        # name). Cloud provider will be e.g. "EC2" indicating that we're running on e.g.
        # an EC2 instance allocated via instance_allocation.py, so we need to deallocate
        # the job via the right InstanceRegistrar when we're done. region name indicates
        # where the InstanceRegistrar that we used to allocate this job is.
        self.cloud_provider = cloud_provider
        self.instance_name = instance_name
        self._connect_task: Optional[asyncio.Task[asyncssh.SSHClientConnection]] = None

    def _connection_future(self) -> asyncio.Task[asyncssh.SSHClientConnection]:
        # try the connection 20 times.
        if self._connect_task is None:
            self._connect_task = asyncio.create_task(
                _retry(
                    lambda: ssh.connect(
                        self.address,
                        username=self.username,
                        private_key=self.private_key,
                    ),
                    (TimeoutError, ConnectionRefusedError, OSError),
                    max_num_attempts=20,
                )
            )
        return self._connect_task

    async def run_job(
        self,
        resources_required: Optional[ResourcesInternal],
        job: Job,
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        # TODO we should probably make SshHost not a Host, and if we want to expose
        # SshHost-as-a-Host functionality, create a separate, more pure SshHost class.
        # This SshHost class is fairly tightly coupled to the EC2/AzureVM code.
        return await self.run_cloud_job(job, wait_for_result, None)

    async def run_cloud_job(
        self,
        job: Job,
        wait_for_result: WaitOption,
        deallocator: Optional[Callable[[], Awaitable[None]]],
    ) -> JobCompletion[Any]:
        job_io_prefix = ""
        deallocation_ran = False

        try:
            connection = await self._connection_future()

            # serialize job_to_run and send it to the remote machine
            job_io_prefix = f"/var/meadowrun/io/{job.job_id}"
            await ssh.write_to_file(
                connection, job.SerializeToString(), f"{job_io_prefix}.job_to_run"
            )

            command_prefixes = []
            command_suffixes = []

            if wait_for_result == WaitOption.DO_NOT_WAIT:
                # reference on nohup: https://github.com/ronf/asyncssh/issues/137
                command_prefixes.append("/usr/bin/nohup")
            if self.cloud_provider is not None:
                command_suffixes.append(
                    f"--cloud {self.cloud_provider[0]} "
                    f"--cloud-region-name {self.cloud_provider[1]}"
                )

            log_file_name = (
                f"/var/meadowrun/job_logs/{job.job_friendly_name}.{job.job_id}.log"
            )
            if wait_for_result == WaitOption.WAIT_AND_TAIL_STDOUT:
                command_suffixes.append(
                    f"2>&1 | tee --ignore-interrupts {log_file_name}"
                )
            else:
                command_suffixes.append(f"> {log_file_name} 2>&1")
                if wait_for_result == WaitOption.DO_NOT_WAIT:
                    command_suffixes.append("&")

            if command_prefixes:
                command_prefixes.append(" ")
            if command_suffixes:
                command_suffixes.insert(0, " ")

            command = (
                " ".join(command_prefixes) + "/usr/bin/env PYTHONUNBUFFERED=1 "
                "/var/meadowrun/env/bin/python "
                # "-X importtime "
                # "-m cProfile -o remote.prof "
                "-m meadowrun.run_job_local_main "
                f"--job-id {job.job_id}" + " ".join(command_suffixes)
            )

            print(f"Running job on {self.address} {log_file_name}")

            # very shortly after this point, we can rely on the deallocation command
            # running on the remote machine
            deallocation_ran = True
            cmd_result = await ssh.run_and_print(connection, command, check=False)

            # TODO consider using result.tail, result.stdout

            # see if we got a normal return code
            if cmd_result.exit_status != 0:
                raise ValueError(f"Process exited {cmd_result.returncode}")

            if wait_for_result == WaitOption.DO_NOT_WAIT:
                return JobCompletion(
                    None, ProcessState.ProcessStateEnum.RUNNING, "", 0, self.address
                )

            process_state = ProcessState()
            process_state.ParseFromString(
                await ssh.read_from_file(connection, f"{job_io_prefix}.process_state")
            )

            if process_state.state == ProcessState.ProcessStateEnum.SUCCEEDED:
                job_spec_type = job.WhichOneof("job_spec")
                # we must have a result from functions, in other cases we can
                # optionally have a result
                if job_spec_type == "py_function" or process_state.pickled_result:
                    result = pickle.loads(process_state.pickled_result)
                else:
                    result = None
                return JobCompletion(
                    result,
                    process_state.state,
                    process_state.log_file_name,
                    process_state.return_code,
                    self.address,
                )
            else:
                raise MeadowrunException(process_state, self.address)

        finally:
            if not deallocation_ran and deallocator:
                await deallocator()

            # TODO also clean up log files?
            # TODO this logic should be moved into the remote machine, a la
            # deallocate_jobs.py. That will also take care of cleaning up these files
            # during DO_NOT_WAIT
            if job_io_prefix and wait_for_result != WaitOption.DO_NOT_WAIT:
                remote_paths = " ".join(
                    [
                        f"{job_io_prefix}.job_to_run",
                        f"{job_io_prefix}.function",
                        f"{job_io_prefix}.arguments",
                        f"{job_io_prefix}.state",
                        f"{job_io_prefix}.result",
                        f"{job_io_prefix}.process_state",
                        f"{job_io_prefix}.initial_process_state",
                    ]
                )
                try:
                    # -f so that we don't throw an error on files that don't
                    # exist
                    await ssh.run_and_capture(
                        connection, f"rm -f {remote_paths}", check=True
                    )
                    pass
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    print(
                        f"Error cleaning up files on remote machine: {remote_paths} {e}"
                    )

    async def close_connection(self) -> None:
        if self._connect_task is not None:
            connection = await self._connect_task
            connection.close()
            await connection.wait_closed()

    async def run_map(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> Optional[Sequence[_U]]:
        raise NotImplementedError("run_map is not implemented for SshHost")

    def run_map_as_completed(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> AsyncIterable[TaskResult[_U]]:
        raise NotImplementedError("run_map_as_completed is not implemented for SshHost")

    async def get_storage_bucket(self) -> AbstractStorageBucket:
        raise NotImplementedError("get_storage_bucket is not implemented for SshHost")


@dataclasses.dataclass
class JobCompletion(Generic[_T]):
    """Information about how a job completed"""

    # TODO both JobCompletion and MeadowrunException should be revisited

    result: _T
    process_state: ProcessState._ProcessStateEnum.ValueType
    log_file_name: str
    return_code: int
    public_address: str


class MeadowrunException(Exception):
    def __init__(
        self, process_state: ProcessState, address: Optional[str] = None
    ) -> None:
        if process_state.state in (
            ProcessState.ProcessStateEnum.NON_ZERO_RETURN_CODE,
            ProcessState.ProcessStateEnum.UNEXPECTED_WORKER_EXIT,
        ):
            return_code = f": {process_state.return_code}"
        else:
            return_code = ""

        message = [
            "Failure while running a Meadowrun job: "
            f"{ProcessState.ProcessStateEnum.Name(process_state.state)}{return_code}"
        ]

        if process_state.log_file_name:
            if address is not None:
                message.append(f"Log file: {address} {process_state.log_file_name}")
            else:
                message.append(f"Log file: {process_state.log_file_name}")
        elif address is not None:
            message.append(f"On host: {address}")

        if (
            process_state.state
            in (
                ProcessState.ProcessStateEnum.RUN_REQUEST_FAILED,
                ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
                ProcessState.ProcessStateEnum.ERROR_GETTING_STATE,
            )
            and process_state.pickled_result
        ):
            remote_exception = unpickle_exception(process_state.pickled_result)
            message.append(remote_exception[2])

        super().__init__("\n".join(message))
        self.process_state = process_state


class TaskException(Exception):
    """Represents an exception that occurred in a task."""

    pass


@dataclasses.dataclass(frozen=True)
class TaskResult(Generic[_T]):
    """
    The result of a [run_map_as_completed][meadowrun.run_map_as_completed] task.

    Attributes:
        task_id: The index of the task as it was originally passed to
            `run_map_as_completed`.
        is_success: True if the task completed successfully, False if the task raised an
            exception
        result: If `is_success`, the result of the task. Otherwise, None. See also
            `result_or_raise`
        exception: If `not is_success`, a Tuple describing the exception that the task
            raised. Otherwise, None. See also `result_or_raise`.
        attempt: 1-based number indicating which attempt of the task this is. 1 means
            first attempt, 2 means second attempt, etc.
    """

    task_id: int
    is_success: bool
    state: str
    result: Optional[_T] = None
    exception: Optional[Tuple[str, str, str]] = None
    attempt: int = 1
    log_file_name: str = ""

    @staticmethod
    def from_process_state(task: TaskProcessState) -> TaskResult:
        if task.result.state == ProcessState.ProcessStateEnum.SUCCEEDED:
            try:
                unpickled_result = pickle.loads(task.result.pickled_result)
            except asyncio.CancelledError:
                raise
            except BaseException as e:
                # the task was successful but now the result cannot be unpickled. This
                # is usually because there is a dependency available on the remote side
                # that is not available locally.
                return TaskResult(
                    task.task_id,
                    is_success=False,
                    state="RESULT_CANNOT_BE_UNPICKLED",
                    exception=(str(type(e)), str(e), traceback.format_exc()),
                    result=task.result.pickled_result,
                    attempt=task.attempt,
                    log_file_name=task.result.log_file_name,
                )

            return TaskResult(
                task.task_id,
                is_success=True,
                state=ProcessState.ProcessStateEnum.Name(task.result.state),
                result=unpickled_result,
                attempt=task.attempt,
                log_file_name=task.result.log_file_name,
            )
        elif task.result.state in _EXCEPTION_STATES:
            exception = unpickle_exception(task.result.pickled_result)
            if exception is None and task.result.return_code != 0:
                return_code_message = f"Non-zero return code: {task.result.return_code}"
                exception = ("", return_code_message, return_code_message + "\n")
            return TaskResult(
                task.task_id,
                is_success=False,
                state=ProcessState.ProcessStateEnum.Name(task.result.state),
                exception=exception,
                attempt=task.attempt,
                log_file_name=task.result.log_file_name,
            )
        else:
            if task.result.return_code != 0:
                return_code_message = f"Non-zero return code: {task.result.return_code}"
                exception = ("", return_code_message, return_code_message + "\n")
            else:
                exception = None
            return TaskResult(
                task.task_id,
                is_success=False,
                state=ProcessState.ProcessStateEnum.Name(task.result.state),
                exception=exception,
                attempt=task.attempt,
                log_file_name=task.result.log_file_name,
            )

    def result_or_raise(self) -> _T:
        """Returns a successful task result, or raises a TaskException.

        Raises:
            TaskException: if the task did not finish successfully.

        Returns:
            _T: the unpickled result if the task finished successfully.
        """
        if self.is_success:
            assert self.result is not None
            return self.result
        elif self.exception is not None:
            raise TaskException(*self.exception)
        else:
            # fallback exception
            raise TaskException(f"Task was not successful, state: {self.state}")


_EXCEPTION_STATES = (
    ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
    ProcessState.ProcessStateEnum.RUN_REQUEST_FAILED,
)


@dataclasses.dataclass(frozen=True)
class ContainerRegistryHelper:
    """
    Allows compile_environment_spec_to_container to use either AWS ECR, Azure CR, or
    neither
    """

    should_push: bool
    username_password: Optional[UsernamePassword]
    image_name: str
    does_image_exist: bool


class RunMapTasksFailedException(Exception):
    def __init__(self, failed_tasks: List[TaskResult], failed_task_args: List[Any]):
        failed_task_message = [f"{len(failed_tasks)} tasks failed:\n"]
        for task, arg in zip(failed_tasks, failed_task_args):
            failed_task_message.append(
                f"Task #{task.task_id} with arg ({arg}) on "
                f"attempt {task.attempt} failed, log file is {task.log_file_name}, "
                f"final state: {task.state}, "
                f"exception was:\n"
            )
            if task.exception is not None:
                failed_task_message.append(task.exception[2])
            else:
                failed_task_message.append("Exception traceback not available\n")

        super().__init__("".join(failed_task_message))

        self.failed_tasks = failed_tasks
        self.failed_task_args = failed_task_args


class AllocVM(Host, abc.ABC):
    """
    An abstract class that provides shared implementation for
    [AllocEC2Instance][meadowrun.AllocEC2Instance] and
    [AllocAzureVM][meadowrun.AllocAzureVM]
    """

    async def run_map_as_completed(
        self,
        function: Callable[[_T], _U],
        args: Sequence[_T],
        resources_required_per_task: Optional[ResourcesInternal],
        job_fields: Dict[str, Any],
        num_concurrent_tasks: int,
        pickle_protocol: int,
        wait_for_result: WaitOption,
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> AsyncIterable[TaskResult[_U]]:
        if resources_required_per_task is None:
            raise ValueError(
                "Resources.logical_cpu and memory_gb must be specified for "
                "AllocEC2Instance and AllocAzureVM"
            )

        async with self._create_grid_job_cloud_interface() as cloud_interface:
            driver = GridJobDriver(
                cloud_interface, num_concurrent_tasks, resources_required_per_task
            )
            run_worker_loops = asyncio.create_task(
                driver.run_worker_functions(
                    self,
                    function,
                    pickle_protocol,
                    job_fields,
                    wait_for_result,
                )
            )
            num_tasks_done = 0
            async for result in driver.add_tasks_and_get_results(
                args, max_num_task_attempts, retry_with_more_memory
            ):
                yield result
                num_tasks_done += 1

            await run_worker_loops

        # this is for extra safety--the only case where we don't get all of our results
        # back should be if run_worker_loops throws an exception because there were
        # worker failures
        if num_tasks_done < len(args):
            raise ValueError(
                "Gave up retrieving task results, most likely due to worker failures. "
                f"Received {num_tasks_done}/{len(args)} task results."
            )

    @abc.abstractmethod
    def _create_grid_job_cloud_interface(self) -> GridJobCloudInterface:
        pass

    @abc.abstractmethod
    def get_cloud_provider(self) -> CloudProviderType:
        pass

    @abc.abstractmethod
    def get_runtime_resources(self) -> ResourcesInternal:
        # "Runtime resources" are resources that aren't tied to a particular instance
        # type. Instance type resources are things like CPU and memory. Runtime
        # resources are things like an AMI id or subnet id. Runtime resources should NOT
        # be considered when choosing an instance type, but need to be considered when
        # deciding whether an existing instance can run a new job.
        pass


@dataclasses.dataclass(frozen=True)
class TaskProcessState:
    task_id: int
    attempt: int
    result: ProcessState


@dataclasses.dataclass(frozen=True)
class WorkerProcessState:
    worker_index: str
    result: ProcessState


class GridJobCloudInterface(abc.ABC, Generic[_T, _U]):
    """
    See also GridJobDriver. The GridJobDriver is a concrete class that handles the
    cloud-independent logic for running a grid job, e.g. replacing workers that exit
    unexpectedly (not implemented yet), retrying vs giving up on tasks. The
    GridJobCloudInterface is an interface that we can implement for different cloud
    providers. The GridJobDriver always has a single GridJobCloudInterface
    implementation that it uses to actually "do things" in the real world like launch
    instances and start workers.
    """

    async def __aenter__(self) -> GridJobCloudInterface:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    @abc.abstractmethod
    def create_instance_registrar(self) -> InstanceRegistrar:
        ...

    @abc.abstractmethod
    def create_queue(self) -> int:
        ...

    @abc.abstractmethod
    async def setup_and_add_tasks(self, tasks: Sequence[_T]) -> None:
        """
        GridJobDriver will always call this exactly once before any other functions on
        this class are called.
        """
        ...

    @abc.abstractmethod
    async def ssh_host_from_address(self, address: str, instance_name: str) -> SshHost:
        ...

    @abc.abstractmethod
    async def shutdown_workers(self, num_workers: int, queue_index: int) -> None:
        ...

    @abc.abstractmethod
    async def get_worker_function(
        self, queue_index: int
    ) -> Callable[
        [str, str, TaskWorkerServer, WorkerMonitor],
        Coroutine[Any, Any, None],
    ]:
        """
        Returns a function that will poll/wait for tasks, and communicate to task worker
        via the given streamreader and -writer. The returned function will also exit in
        response to shutdown_workers.
        """
        ...

    @abc.abstractmethod
    async def receive_task_results(
        self, *, stop_receiving: asyncio.Event, workers_done: asyncio.Event
    ) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
        ...

    @abc.abstractmethod
    async def retry_task(
        self, task_id: int, attempts_so_far: int, queue_index: int
    ) -> None:
        ...


_PRINT_RECEIVED_TASKS_SECONDS = 10


@dataclasses.dataclass
class WorkerQueue:
    """
    Keeps track of how many workers we've started/shutdown for each queue. The 0th queue
    is the "normal" one, and subsequent queues are for retrying tasks with more
    resources
    """

    queue_index: int
    num_workers_needed: int
    num_workers_launched: int = 0
    num_worker_shutdown_messages_sent: int = 0
    num_workers_exited_unexpectedly: int = 0
    pickled_worker_function_task: Optional[asyncio.Task[bytes]] = None


@dataclasses.dataclass(frozen=True)
class WorkerTask:
    log_file_info: str
    queue_index: int
    task: asyncio.Task[JobCompletion]


def _memory_gb_for_queue_index(
    queue_index: int, base_resources: ResourcesInternal
) -> float:
    """See WorkerQueue"""
    return base_resources.consumable[MEMORY_GB] * (1 + queue_index)


def _resources_for_queue_index(
    queue_index: int, base_resources: ResourcesInternal
) -> ResourcesInternal:
    """See WorkerQueue"""
    if queue_index == 0:
        return base_resources
    else:
        return base_resources.add(
            ResourcesInternal.from_cpu_and_memory(
                0,
                base_resources.consumable[MEMORY_GB] * queue_index,
            )
        )


class GridJobDriver:
    """
    See GridJobCloudInterface. This class handles the cloud-independent logic for
    running a grid job, e.g. replacing workers that exit unexpectedly (not implemented
    yet), retrying vs giving up on tasks.

    The basic design is that there are two "loops" run_worker_functions and
    add_tasks_and_get_results that run "in parallel" via asyncio. They interact using
    shared asyncio.Event objects.
    """

    def __init__(
        self,
        cloud_interface: GridJobCloudInterface,
        num_concurrent_tasks: int,
        resources_required_per_task: ResourcesInternal,
    ):
        """This constructor must be called on an EventLoop"""
        self._cloud_interface = cloud_interface
        self._resources_required_per_task = resources_required_per_task

        # run_worker_functions will set this to indicate to add_tasks_and_get_results
        # that there all of our workers have either exited unexpectedly (and we have
        # given up trying to restore them), or have been told to shutdown normally
        self._no_workers_available = asyncio.Event()

        self._worker_queues = [WorkerQueue(0, num_concurrent_tasks)]
        self._num_workers_needed_changed = asyncio.Event()

        self._abort_launching_new_workers = asyncio.Event()

    async def run_worker_functions(
        self,
        alloc_cloud_instance: AllocVM,
        user_function: Callable[[_T], _U],
        pickle_protocol: int,
        job_fields: Dict[str, Any],
        wait_for_result: WaitOption,
    ) -> None:
        """
        Allocates cloud instances, runs a worker function on them, sends worker shutdown
        messages when requested by add_tasks_and_get_results, and generally manages
        workers (e.g. replacing workers when they exit unexpectedly, not implemented
        yet).
        """

        address_to_ssh_host: Dict[str, SshHost] = {}
        worker_tasks: List[WorkerTask] = []

        workers_needed_changed_wait_task = asyncio.create_task(
            self._num_workers_needed_changed.wait()
        )
        pickled_worker_function_task: Optional[asyncio.Task[bytes]] = None
        async_cancel_exception = False

        try:
            if (
                job_fields["ports"]
                and alloc_cloud_instance.get_cloud_provider() == "AzureVM"
            ):
                raise NotImplementedError(
                    "Opening ports on Azure is not implemented, please comment on "
                    "https://github.com/meadowdata/meadowrun/issues/126"
                )

            # we create an asyncio.task for this because it requires waiting for
            # _cloud_interface.setup_and_add_tasks to complete. We don't want to just
            # run this sequentially, though, because we want to start launching
            # instances before setup_and_add_tasks is complete.
            async def _get_pickled_worker_function(queue_index: int) -> bytes:
                return cloudpickle.dumps(
                    await self._cloud_interface.get_worker_function(queue_index),
                    protocol=pickle_protocol,
                )

            # initially, we just want the 0th queue
            self._worker_queues[0].pickled_worker_function_task = asyncio.create_task(
                _get_pickled_worker_function(0)
            )

            # "inner_" on the parameter names is just to avoid name collision with outer
            # scope
            async def launch_worker_function(
                inner_ssh_host: SshHost,
                inner_worker_job_id: str,
                log_file_name: str,
                inner_instance_registrar: InstanceRegistrar,
                queue_index: int,
            ) -> JobCompletion:
                inner_worker_queue = self._worker_queues[queue_index]
                if inner_worker_queue.pickled_worker_function_task is None:
                    inner_worker_queue.pickled_worker_function_task = (
                        asyncio.create_task(_get_pickled_worker_function(queue_index))
                    )
                job = Job(
                    job_id=inner_worker_job_id,
                    py_agent=PyAgentJob(
                        pickled_function=cloudpickle.dumps(
                            user_function, protocol=pickle_protocol
                        ),
                        pickled_agent_function=(
                            await inner_worker_queue.pickled_worker_function_task
                        ),
                        pickled_agent_function_arguments=pickle.dumps(
                            ([inner_ssh_host.address, log_file_name], {}),
                            protocol=pickle_protocol,
                        ),
                    ),
                    **job_fields,
                )

                async def deallocator() -> None:
                    if inner_ssh_host.instance_name is not None:
                        await inner_instance_registrar.deallocate_job_from_instance(
                            await inner_instance_registrar.get_registered_instance(
                                inner_ssh_host.instance_name
                            ),
                            inner_worker_job_id,
                        )

                return await inner_ssh_host.run_cloud_job(
                    job, wait_for_result, deallocator
                )

            async with self._cloud_interface.create_instance_registrar() as instance_registrar:  # noqa: E501
                while True:
                    for worker_queue in self._worker_queues:
                        # TODO we should subtract workers_exited_unexpectedly from
                        # workers_launched, this would mean we replace workers that
                        # exited unexpectedly. Implementing this properly means we
                        # should add more code that will tell us why workers exited
                        # unexpectedly (e.g. segfault in user code, spot instance
                        # eviction, vs an issue creating the environment). The main
                        # concern is ending up in an infinite loop where we're
                        # constantly launching workers that fail.
                        new_workers_to_launch = worker_queue.num_workers_needed - (
                            worker_queue.num_workers_launched
                            - worker_queue.num_worker_shutdown_messages_sent
                        )

                        # launch new workers if they're needed
                        if new_workers_to_launch > 0:
                            # TODO pass in an event for if workers_needed goes to 0,
                            # cancel any remaining allocate_jobs_to_instances
                            async for allocated_hosts in allocate_jobs_to_instances(
                                instance_registrar,
                                _resources_for_queue_index(
                                    worker_queue.queue_index,
                                    self._resources_required_per_task,
                                ),
                                new_workers_to_launch,
                                alloc_cloud_instance,
                                job_fields["ports"],
                                self._abort_launching_new_workers,
                            ):
                                for (
                                    (public_address, instance_name),
                                    worker_job_ids,
                                ) in allocated_hosts.items():
                                    ssh_host = address_to_ssh_host.get(public_address)
                                    if ssh_host is None:
                                        ssh_host = await self._cloud_interface.ssh_host_from_address(  # noqa: E501
                                            public_address, instance_name
                                        )
                                        address_to_ssh_host[public_address] = ssh_host
                                    for worker_job_id in worker_job_ids:
                                        log_file_name = (
                                            "/var/meadowrun/job_logs/"
                                            f"{job_fields['job_friendly_name']}."
                                            f"{worker_job_id}.log"
                                        )
                                        worker_tasks.append(
                                            WorkerTask(
                                                f"{public_address} {log_file_name}",
                                                worker_queue.queue_index,
                                                asyncio.create_task(
                                                    launch_worker_function(
                                                        ssh_host,
                                                        worker_job_id,
                                                        log_file_name,
                                                        instance_registrar,
                                                        worker_queue.queue_index,
                                                    )
                                                ),
                                            )
                                        )
                                        worker_queue.num_workers_launched += 1

                        # shutdown workers if they're no longer needed
                        # once we have the lost worker replacement logic this should
                        # just be -workers_to_launch
                        workers_to_shutdown = (
                            worker_queue.num_workers_launched
                            - worker_queue.num_worker_shutdown_messages_sent
                            - worker_queue.num_workers_exited_unexpectedly
                        ) - (worker_queue.num_workers_needed)
                        if workers_to_shutdown > 0:
                            await self._cloud_interface.shutdown_workers(
                                workers_to_shutdown, worker_queue.queue_index
                            )
                            worker_queue.num_worker_shutdown_messages_sent += (
                                workers_to_shutdown
                            )

                    # now we wait until either add_tasks_and_get_results tells us to
                    # shutdown some workers, or workers exit
                    await asyncio.wait(
                        itertools.chain(
                            cast(
                                Iterable[asyncio.Task],
                                (task.task for task in worker_tasks),
                            ),
                            (workers_needed_changed_wait_task,),
                        ),
                        return_when=asyncio.FIRST_COMPLETED,
                    )
                    new_worker_tasks = []
                    for worker_task in worker_tasks:
                        if worker_task.task.done():
                            exception = worker_task.task.exception()
                            if exception is not None:
                                print(
                                    # TODO also include queue index here?
                                    f"Error running worker {worker_task.log_file_info}:"
                                    f"\n"
                                    + "".join(
                                        traceback.format_exception(
                                            type(exception),
                                            exception,
                                            exception.__traceback__,
                                        )
                                    )
                                )
                                self._worker_queues[
                                    worker_task.queue_index
                                ].num_workers_exited_unexpectedly += 1
                                # TODO ideally we would tell the receive_results loop to
                                # reschedule whatever task the worker was working on
                            #  TODO do something with worker_task.result()
                        else:
                            new_worker_tasks.append(worker_task)
                    worker_tasks = new_worker_tasks

                    if workers_needed_changed_wait_task.done():
                        self._num_workers_needed_changed.clear()
                        workers_needed_changed_wait_task = asyncio.create_task(
                            self._num_workers_needed_changed.wait()
                        )

                    # this means all workers are either done or shutting down
                    if all(
                        worker_queue.num_workers_launched
                        - worker_queue.num_worker_shutdown_messages_sent
                        - worker_queue.num_workers_exited_unexpectedly
                        <= 0
                        for worker_queue in self._worker_queues
                    ):
                        break

        except asyncio.CancelledError:
            # if we're being cancelled, then most likely the worker_tasks are being
            # cancelled as well (because someone pressed Ctrl+C), which means it's
            # unhelpful to cancel them again, we want them to finish running their
            # except/finally clauses
            async_cancel_exception = True
            raise
        finally:
            print("Shutting down workers")
            # setting this is very critical--otherwise, add_tasks_and_get_results will
            # hang forever, not knowing that it has no hope of workers working on any of
            # its tasks
            self._no_workers_available.set()

            if not async_cancel_exception:
                # cancel any outstanding workers. Even if there haven't been any
                # exception and we've sent worker shutdown messages for all the workers,
                # a worker might be in the middle of building/pulling an environment and
                # we don't want to wait for that to complete
                for worker_task in worker_tasks:
                    worker_task.task.cancel()

            # we still want the tasks to complete their except/finally blocks, after
            # which they should reraise asyncio.CancelledError, which we can safely
            # ignore
            await asyncio.gather(
                *(task.task for task in worker_tasks), return_exceptions=True
            )

            await asyncio.gather(
                *[
                    ssh_host.close_connection()
                    for ssh_host in address_to_ssh_host.values()
                ],
                return_exceptions=True,
            )
            workers_needed_changed_wait_task.cancel()
            if pickled_worker_function_task is not None:
                pickled_worker_function_task.cancel()

    async def add_tasks_and_get_results(
        self,
        args: Sequence[_T],
        max_num_task_attempts: int,
        retry_with_more_memory: bool,
    ) -> AsyncIterable[TaskResult]:
        """
        Adds the specified tasks to the "queue", and retries tasks as needed. Yields
        TaskResult objects as soon as tasks complete.
        """

        # this keeps track of which queue each arg is assigned to. Initially they are
        # all assigned to the 0th queue, and if they fail because of suspected lack of
        # memory and retry_with_more_memory is set, then we will increase their queue
        # index. -1 indicates that the arg is done so it is no longer assigned to any
        # queue
        arg_to_queue_index = [0] * len(args)
        await self._cloud_interface.setup_and_add_tasks(args)

        # done = successful or exhausted retries
        num_tasks_done = 0
        # stop_receiving tells _cloud_interface.receive_task_results that there are no
        # more results to get
        stop_receiving = asyncio.Event()
        if len(args) == num_tasks_done:
            stop_receiving.set()
        last_printed_update = time.time()
        print(
            f"Waiting for task results. Requested: {len(args)}, "
            f"Done: {num_tasks_done}"
        )
        async for task_batch, worker_batch in await self._cloud_interface.receive_task_results(  # noqa: E501
            stop_receiving=stop_receiving, workers_done=self._no_workers_available
        ):
            # TODO right now we ignore worker_batch because we get worker failures
            # through the SSH connection. At some point, we may want to process worker
            # failures here.

            for task in task_batch:
                task_result = TaskResult.from_process_state(task)
                if task_result.is_success:
                    num_tasks_done += 1
                    arg_to_queue_index[task.task_id] = -1
                    yield task_result
                elif (
                    task.attempt < max_num_task_attempts
                    and task_result.state != "RESULT_CANNOT_BE_UNPICKLED"
                ):
                    prev_queue_index = arg_to_queue_index[task.task_id]
                    prev_memory_requirement = _memory_gb_for_queue_index(
                        prev_queue_index, self._resources_required_per_task
                    )
                    if (
                        retry_with_more_memory
                        and task.result.max_memory_used_gb
                        >= 0.95 * prev_memory_requirement
                    ):
                        print(
                            f"Task {task.task_id} failed at attempt {task.attempt}, "
                            "retrying with more memory (task used "
                            f"{task.result.max_memory_used_gb:.2f}/"
                            f"{prev_memory_requirement}GB requested)"
                        )

                        new_queue_index = arg_to_queue_index[task.task_id] + 1
                        arg_to_queue_index[task.task_id] = new_queue_index
                        if len(self._worker_queues) < new_queue_index + 1:
                            # TODO any new queue gets 1 worker by default. We should
                            # increase this later depending on how the situation changes
                            self._cloud_interface.create_queue()
                            self._worker_queues.append(WorkerQueue(new_queue_index, 1))
                        await self._cloud_interface.retry_task(
                            task.task_id, task.attempt, new_queue_index
                        )
                    else:
                        print(
                            f"Task {task.task_id} failed at attempt {task.attempt}, "
                            "retrying"
                        )
                        await self._cloud_interface.retry_task(
                            task.task_id, task.attempt, prev_queue_index
                        )
                else:
                    if (
                        task.attempt < max_num_task_attempts
                        and task_result.state == "RESULT_CANNOT_BE_UNPICKLED"
                    ):
                        print(
                            f"Task {task.task_id} failed at attempt {task.attempt}, max"
                            f" attempts is {max_num_task_attempts}, but not retrying "
                            "because the failure happened when trying to unpickle the "
                            "result on the client."
                        )
                    else:
                        print(
                            f"Task {task.task_id} failed at attempt {task.attempt}, max"
                            f" attempts is {max_num_task_attempts}, not retrying."
                        )
                    num_tasks_done += 1
                    arg_to_queue_index[task.task_id] = -1
                    yield task_result

            if num_tasks_done >= len(args):
                stop_receiving.set()
            else:
                t0 = time.time()
                if t0 - last_printed_update > _PRINT_RECEIVED_TASKS_SECONDS:
                    print(
                        f"Waiting for task results. Requested: {len(args)}, "
                        f"Done: {num_tasks_done}"
                    )
                    last_printed_update = t0

            # reduce the number of workers needed if we have more workers than
            # outstanding tasks
            num_workers_needed_per_queue = collections.Counter(arg_to_queue_index)
            for worker_queue in self._worker_queues:
                # num_workers_needed = min(outstanding tasks for this queue, current num
                # workers needed for this queue)
                # TODO at some point we might want to increase the number of workers
                # needed?
                num_workers_needed = min(
                    num_workers_needed_per_queue.get(worker_queue.queue_index, 0),
                    worker_queue.num_workers_needed,
                )
                if num_workers_needed < worker_queue.num_workers_needed:
                    worker_queue.num_workers_needed = num_workers_needed
                    self._num_workers_needed_changed.set()

        # We could be more finegrained about aborting launching workers. This is the
        # easiest to implement, but ideally every time num_workers_needed changes we
        # would consider cancelling launching new workers
        self._abort_launching_new_workers.set()

        if num_tasks_done < len(args):
            # It would make sense for this to raise an exception, but it's more helpful
            # to see the actual worker failures, and run_worker_functions should always
            # raise an exception in that case. The caller should still check though that
            # we returned all of the task results we were expecting.
            print(
                "Gave up retrieving task results, most likely due to worker failures. "
                f"Received {num_tasks_done}/{len(args)} task results."
            )
        else:
            print(f"Received all {len(args)} task results.")



# Copied from https://pypi.org/project/aiodocker/0.21.0/
# Requirements are python >= 3.6 aiohttp >= 3.6, typing-extensions >= 3.6.5
# There's currently no conda package for this so we vendorize it

from .docker import Docker
from .exceptions import DockerContainerError, DockerError


__version__ = "0.21.0"


__all__ = ("Docker", "DockerError", "DockerContainerError")



"""
This code belongs in either func_worker_storage.py. Unfortunately, if we put it in
func_worker_storage.py, then meadowrun.__init__ will import k8s_integration.k8s which
will import func_worker_storage, which means that running func_worker_storage will
produce annoying messages like RuntimeWarning: 'meadowrun.func_worker_storage' found in
sys.modules after import of package 'meadowrun', but prior to execution of
'meadowrun.func_worker_storage'; this may result in unpredictable behaviour
warn(RuntimeWarning(msg))
"""
from __future__ import annotations

from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from meadowrun.abstract_storage_bucket import AbstractStorageBucket

MEADOWRUN_STORAGE_USERNAME = "MEADOWRUN_STORAGE_USERNAME"
MEADOWRUN_STORAGE_PASSWORD = "MEADOWRUN_STORAGE_PASSWORD"

# This is a global variable that will be updated with the storage client if it's
# available in func_worker_storage
FUNC_WORKER_STORAGE_BUCKET: Optional[AbstractStorageBucket] = None



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['meadowrun',
 'meadowrun._vendor',
 'meadowrun._vendor.aiodocker',
 'meadowrun._vendor.fastcdc',
 'meadowrun._vendor.gcloud',
 'meadowrun._vendor.gcloud.aio',
 'meadowrun._vendor.gcloud.aio.auth',
 'meadowrun._vendor.gcloud.aio.storage',
 'meadowrun._vendor.platformdirs',
 'meadowrun.aws_integration',
 'meadowrun.aws_integration.management_lambdas',
 'meadowrun.azure_integration',
 'meadowrun.azure_integration.mgmt_functions',
 'meadowrun.azure_integration.mgmt_functions.azure_core',
 'meadowrun.azure_integration.mgmt_functions.clean_up',
 'meadowrun.azure_integration.mgmt_functions.vm_adjust',
 'meadowrun.deployment',
 'meadowrun.func_worker',
 'meadowrun.gcp_integration',
 'meadowrun.k8s_integration']

package_data = \
{'': ['*'], 'meadowrun': ['docker_files/*']}

install_requires = \
['aiobotocore>=2.1.2,<3.0.0',
 'aiofiles>=0.6.0,<0.7.0',
 'aiohttp>=3.8.0,<4.0.0',
 'asyncssh>=2.11.0,<3.0.0',
 'backoff>=1.0.0,<3.0.0',
 'boto3>=1.21.21,<2.0.0',
 'chardet>=2.0,<4.1',
 'cloudpickle>=2.0.0,<3.0.0',
 'cryptography>=2.0.0,<37.0.0',
 'filelock>=3.6.0,<4.0.0',
 'future>=0.17.0,<0.18.3',
 'kubernetes-asyncio>=24.2.0,<25.0.0',
 'protobuf>=3.18.1,<4.0.0',
 'psutil>=5.8.0,<6.0.0',
 'pyasn1-modules==0.2.1',
 'pyjwt>=1.5.3,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'rsa>=3.1.4,<5.0.0',
 'six>=1.11.0,<2.0.0',
 'typing-extensions>=4.1.1,<5.0.0']

entry_points = \
{'console_scripts': ['meadowrun-local = '
                     'meadowrun.run_job_local_main:command_line_main',
                     'meadowrun-manage-azure-vm = '
                     'meadowrun.manage:main_azure_vm',
                     'meadowrun-manage-ec2 = meadowrun.manage:main_ec2']}

setup_kwargs = {
    'name': 'meadowrun',
    'version': '0.2.12',
    'description': 'The easiest way to run python code on one or more remote machines',
    'long_description': "# ![Meadowrun](meadowrun-logo-full.svg)\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/meadowrun) ![PyPI](https://img.shields.io/pypi/v/meadowrun)  ![PyPI - Downloads](https://img.shields.io/pypi/dm/meadowrun) ![Conda](https://img.shields.io/conda/v/meadowdata/meadowrun) ![Conda](https://img.shields.io/conda/dn/meadowdata/meadowrun?label=conda%20downloads)\n\n[![Join the chat at https://gitter.im/meadowdata/meadowrun](https://badges.gitter.im/meadowdata/meadowrun.svg)](https://gitter.im/meadowdata/meadowrun?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n\n\n\nMeadowrun is a library for data scientists and data engineers who run python code on\nAWS, Azure, or Kubernetes. Meadowrun:\n\n- scales from a single function to thousands of distributed tasks.\n- syncs your local code and libraries for a faster, easier iteration loop. Edit your\n  code and rerun your analysis without worrying about building packages or Docker\n  images.\n- optimizes for cost, choosing the cheapest instance types and turning them off when\n  they're no longer needed.\n  \nFor more context, see our [case\nstudes](https://docs.meadowrun.io/en/stable/case_studies/) of how Meadowrun is used in\nreal life, or see the [project homepage](https://meadowrun.io)\n\nTo get started, go to our [documentation](https://docs.meadowrun.io), or [join the chat\non Gitter](https://gitter.im/meadowdata/meadowrun)\n\n## Quickstart\n\nFirst, install Meadowrun using pip:\n\n```\npip install meadowrun\n```\n\nNext, assuming you've [configured the AWS\nCLI](https://docs.aws.amazon.com/cli/latest/userguide/getting-started-quickstart.html)\nand are a root/administrator user, you can run:\n\n```python\nimport meadowrun\nimport asyncio\n\nprint(\n    asyncio.run(\n        meadowrun.run_function(\n            lambda: sum(range(1000)) / 1000,\n            meadowrun.AllocEC2Instance(),\n            meadowrun.Resources(logical_cpu=1, memory_gb=8, max_eviction_rate=80),\n            meadowrun.Deployment.mirror_local()\n        )\n    )\n)\n```\n\n[The documentation](https://docs.meadowrun.io) has examples of how to use other package\nmanagers (conda, poetry), and other platforms (Azure, GKE, Kubernetes).\n",
    'author': 'Richard Lee',
    'author_email': 'hrichardlee@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/meadowdata/meadowrun',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Awaitable

if TYPE_CHECKING:
    from typing_extensions import Final

from meadowrun.azure_integration import blob_storage
from meadowrun.azure_integration.azure_mgmt_functions_setup import (
    _create_or_update_mgmt_function_app,
    _update_mgmt_function_code,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    MEADOWRUN_RESOURCE_GROUP_NAME,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    ResourceNotFoundError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_poll,
    get_subscription_id,
    wait_for_poll,
)

RESOURCE_PROVIDERS: Final = (
    "Microsoft.OperationalInsights",
    "microsoft.insights",
    "Microsoft.Storage",
    "Microsoft.ManagedIdentity",
    "Microsoft.KeyVault",
    "Microsoft.Network",
    "Microsoft.Compute",
    "Microsoft.ContainerRegistry",
    "Microsoft.Web",
)


def _ensure_resource_provider_registered(
    subscription_id: str, resource_provider_ns: str
) -> Awaitable[None]:
    """Ensure the given resource provider is registered."""
    return azure_rest_api(
        "POST",
        f"subscriptions/{subscription_id}/providers/{resource_provider_ns}/register",
        "2021-04-01",
    )


async def _register_resource_providers(subscription_id: str) -> None:
    """Register all the resource providers we're using."""

    tasks = [
        _ensure_resource_provider_registered(subscription_id, rp)
        for rp in RESOURCE_PROVIDERS
    ]
    await asyncio.gather(*tasks)


async def delete_meadowrun_resource_group() -> None:
    """
    This should delete all meadowrun-generated resources as deletes everything in the
    meadowrun resource group.
    """
    try:
        # https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/delete
        await wait_for_poll(
            await azure_rest_api_poll(
                "DELETE",
                f"subscriptions/{await get_subscription_id()}/resourcegroups/"
                f"{MEADOWRUN_RESOURCE_GROUP_NAME}",
                "2021-04-01",
                "LocationStatusCode",
            )
        )
    except ResourceNotFoundError:
        pass


async def install(location: str) -> None:
    subscription_id = await get_subscription_id()
    await _register_resource_providers(subscription_id)
    storage_account = await _create_or_update_mgmt_function_app(
        location, subscription_id
    )
    await blob_storage.ensure_container(subscription_id, storage_account, location)
    await _update_mgmt_function_code()



from __future__ import annotations

import asyncio
import io
import os
import time
import zipfile
from typing import TYPE_CHECKING, Tuple, Callable, Awaitable, TypeVar, Union, Type

import aiohttp

import meadowrun.azure_integration.mgmt_functions
from meadowrun.azure_integration.azure_meadowrun_core import (
    _ensure_managed_identity,
    ensure_meadowrun_resource_group,
    ensure_meadowrun_storage_account,
    get_subscription_id,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_identity import (
    get_token,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_poll,
    wait_for_poll,
)
from meadowrun.shared import create_zipfile

if TYPE_CHECKING:
    from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (  # noqa: E501
        StorageAccount,
    )
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE,
    MEADOWRUN_STORAGE_ACCOUNT_VARIABLE,
    MEADOWRUN_SUBSCRIPTION_ID,
)


_T = TypeVar("_T")

_MEADOWRUN_MGMT_FUNCTIONS_LOGS_COMPONENT_NAME = "meadowrun-mgmt-functions-logs"
_MEADOWRUN_MGMT_FUNCTIONS_LOGS_WORKSPACE = "meadowrun-mgmt-functions-logs-workspace"


async def _create_application_insights_component(location: str) -> Tuple[str, str]:
    """
    The Azure Function logs get sent to an application insights component. Returns
    instrumentation key, connection string which is part of the Azure Function
    configuration
    """

    resource_group_path = await ensure_meadowrun_resource_group(location)

    # first, we need to create a log analytics workspace
    workspace_path = (
        f"{resource_group_path}/providers/Microsoft.OperationalInsights/workspaces/"
        f"{_MEADOWRUN_MGMT_FUNCTIONS_LOGS_WORKSPACE}"
    )
    await wait_for_poll(
        await azure_rest_api_poll(
            "PUT",
            workspace_path,
            "2021-06-01",
            "GetProvisioningState",
            json_content={"location": location},
        )
    )

    # then, we can create a application insights component based on that log analytics
    # workspace
    component = await azure_rest_api(
        "PUT",
        (
            f"{resource_group_path}/providers/Microsoft.Insights/components/"
            f"{_MEADOWRUN_MGMT_FUNCTIONS_LOGS_COMPONENT_NAME}"
        ),
        "2020-02-02-preview",
        json_content={
            "location": location,
            "kind": "web",
            "properties": {
                "Application_Type": "web",
                "IngestionMode": "LogAnalytics",
                "WorkspaceResourceId": "/" + workspace_path,
            },
        },
    )
    return (
        component["properties"]["InstrumentationKey"],
        component["properties"]["ConnectionString"],
    )


def _meadowrun_mgmt_function_app_name(subscription_id: str) -> str:
    # This needs to be globally unique
    return "mr" + subscription_id.replace("-", "")


async def _create_or_update_mgmt_function_app(
    location: str, subscription_id: str
) -> StorageAccount:
    """ "Function app" and "site" seem to be interchangeable terms"""
    resource_group_path = await ensure_meadowrun_resource_group(location)

    site_path = (
        f"{resource_group_path}/providers/Microsoft.Web/sites/"
        f"{_meadowrun_mgmt_function_app_name(subscription_id)}"
    )

    # first create some prerequisites:

    # create application insights component for logging
    app_insights_component_task = asyncio.create_task(
        _create_application_insights_component(location)
    )

    # ensure storage account which is required by the functions runtime
    storage_account_task = ensure_meadowrun_storage_account(location, "create")

    # managed identity for the function
    identity_id, identity_client_id = await _ensure_managed_identity(location)

    # now create the actual "function app"/"site"
    # https://docs.microsoft.com/en-us/rest/api/appservice/web-apps/create-or-update
    # oddly, the python SDK thinks that we should be doing the equivalent of an
    # azure_rest_api_poll(poll_scheme="GetProvisioningState"), but there's no
    # "provisioningState" property to check. This seems like a bug/outdated
    # implementation in the SDK.
    await azure_rest_api(
        "PUT",
        site_path,
        "2021-03-01",
        json_content={
            "kind": "functionapp,linux",
            "location": location,
            "identity": {
                "type": "UserAssigned",
                "userAssignedIdentities": {identity_id: {}},
            },
            "properties": {
                # a truly insane name for this parameter, controls Linux vs Windows
                # https://docs.microsoft.com/en-us/azure/templates/microsoft.web/2018-02-01/serverfarms?tabs=bicep#appserviceplanproperties
                "reserved": True,
                "siteConfig": {"linuxFxVersion": "Python|3.9"},
            },
        },
    )

    storage_account = await storage_account_task
    instrumentation_key, connection_string = await app_insights_component_task

    # the "Application Configuration" is a key part of setting up the function app.
    # These settings are technically just environment variables, but they are important
    # for configuring the function runtime
    # https://docs.microsoft.com/en-us/azure/azure-functions/functions-app-settings
    await azure_rest_api(
        "PUT",
        f"{site_path}/config/appsettings",
        "2021-03-01",
        json_content={
            "properties": {
                # these seem to be needed by the function runtime
                "FUNCTIONS_EXTENSION_VERSION": "~4",
                "FUNCTIONS_WORKER_RUNTIME": "python",
                # this is required by the function runtime for its storage
                "AzureWebJobsStorage": (
                    "DefaultEndpointsProtocol=https;"
                    f"AccountName={storage_account.name};"
                    f"AccountKey={storage_account.key};"
                    "EndpointSuffix=core.windows.net"
                ),
                # these variables are needed to for logging to Application Insights
                "APPINSIGHTS_INSTRUMENTATIONKEY": instrumentation_key,
                "APPLICATIONINSIGHTS_CONNECTION_STRING": connection_string,
                # the ManagedIdentityClient seems to need this when running in a
                # function
                "AZURE_CLIENT_ID": identity_client_id,
                # These are used by meadowrun code as environment variables to
                # access the instance registrar table
                MEADOWRUN_STORAGE_ACCOUNT_VARIABLE: storage_account.name,
                MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE: storage_account.key,
                MEADOWRUN_SUBSCRIPTION_ID: subscription_id,
            }
        },
    )
    return storage_account


_IGNORE_FOR_ZIP = {".git", ".venv", ".vscode"}


def _zip_azure_mgmt_function_code() -> bytes:
    """Zips all the code for meadowrun.azure_integration.mgmt_functions"""
    # https://docs.microsoft.com/en-us/azure/azure-functions/functions-reference-python?tabs=asgi%2Cazurecli-linux%2Capplication-level#folder-structure

    zip_path = meadowrun.azure_integration.mgmt_functions.__path__[0]
    first_root_len = None

    with io.BytesIO() as buffer:
        with create_zipfile(buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            for root, dirs, files in os.walk(zip_path):
                to_remove = [d for d in dirs if d in _IGNORE_FOR_ZIP]
                for d in to_remove:
                    # this prevents iteration into these directories
                    dirs.remove(d)

                if first_root_len is None:
                    first_root_len = len(root)

                for file in files:
                    if file not in _IGNORE_FOR_ZIP:
                        zf.write(
                            os.path.join(root, file),
                            os.path.join(root[first_root_len:], file),
                        )

        buffer.seek(0)

        return buffer.read()


async def _update_mgmt_function_code(total_timeout_secs: float = 300) -> None:
    """Assumes the function app already exists, uploads the code for the function app"""
    # https://github.com/Azure/azure-cli/blob/ccdc56e7806b6544ddf228bb04be83e485a7611a/src/azure-cli/azure/cli/command_modules/appservice/custom.py#L498
    # https://docs.microsoft.com/en-us/azure/azure-functions/deployment-zip-push#with-curl

    headers = {"Authorization": f"Bearer {await get_token()}"}

    # this is effectively the same code as azure_rest_api_poll, but we don't have an
    # api-version parameter

    async with aiohttp.request(
        "POST",
        f"https://{_meadowrun_mgmt_function_app_name(await get_subscription_id())}"
        ".scm.azurewebsites.net/api/zipdeploy",
        params={"isAsync": "true"},
        data=_zip_azure_mgmt_function_code(),
        headers=headers,
    ) as response:
        response.raise_for_status()

        if response.status not in (201, 202):
            return

        poll_url = response.headers["Location"]

    t0 = time.time()
    while time.time() - t0 < total_timeout_secs:
        print(f"Waiting for a long-running Azure operation ({5}s)")
        await asyncio.sleep(5)
        async with aiohttp.request("GET", poll_url, headers=headers) as response:
            response.raise_for_status()

            if response.status not in (201, 202):
                return

    raise TimeoutError(
        f"_update_mgmt_function_code timed out after {total_timeout_secs}"
    )


async def _retry(
    function: Callable[[], Awaitable[_T]],
    exception_types: Union[Type, Tuple[Type, ...]],
    max_num_attempts: int = 5,
    delay_seconds: float = 1,
    retry_message: str = "Retrying on error",
) -> _T:
    i = 0
    while True:
        try:
            return await function()
        except exception_types as e:
            i += 1
            if i >= max_num_attempts:
                raise
            else:
                print(f"{retry_message}: {e}")
                await asyncio.sleep(delay_seconds)



from __future__ import annotations

import asyncio
import dataclasses
import datetime
import decimal
import functools
import itertools
import uuid
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import aiobotocore.session
import boto3
from meadowrun.aws_integration.aws_core import _get_default_region_name
from meadowrun.aws_integration.aws_permissions_install import _EC2_ROLE_INSTANCE_PROFILE
from meadowrun.aws_integration.ec2 import (
    LaunchEC2InstanceSettings,
    authorize_current_ip_for_meadowrun_ssh,
    ensure_port_security_group,
    get_ssh_security_group_id,
    launch_ec2_instances,
)
from meadowrun.aws_integration.ec2_ssh_keys import (
    MEADOWRUN_KEY_PAIR_NAME,
    get_meadowrun_ssh_key,
)
from meadowrun.aws_integration.grid_tasks_sqs import (
    add_tasks,
    add_worker_shutdown_messages,
    create_request_queue,
    retry_task,
    worker_function,
)
from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _ALLOCATED_TIME,
    _EC2_ALLOC_TABLE_NAME,
    _EC2_ALLOC_TAG,
    _EC2_ALLOC_TAG_VALUE,
    _INSTANCE_ID,
    _LAST_UPDATE_TIME,
    _NON_CONSUMABLE_RESOURCES,
    _PREVENT_FURTHER_ALLOCATION,
    _PUBLIC_ADDRESS,
    _RESOURCES_ALLOCATED,
    _RESOURCES_AVAILABLE,
    _RUNNING_JOBS,
    ignore_boto3_error_code,
)
from meadowrun.instance_allocation import (
    InstanceRegistrar,
    _InstanceState,
    _TInstanceState,
    allocate_single_job_to_instance,
)
from meadowrun.instance_selection import CloudInstance, ResourcesInternal
from meadowrun.run_job_core import (
    AllocVM,
    CloudProviderType,
    GridJobCloudInterface,
    JobCompletion,
    SshHost,
    WaitOption,
)
from meadowrun.storage_grid_job import (
    S3Bucket,
    get_aws_s3_bucket,
    receive_results,
)

if TYPE_CHECKING:
    from types import TracebackType

    from meadowrun.abstract_storage_bucket import AbstractStorageBucket
    from meadowrun.meadowrun_pb2 import Job
    from meadowrun.run_job_core import TaskProcessState, WorkerProcessState
    from meadowrun.run_job_local import TaskWorkerServer, WorkerMonitor
    from types_aiobotocore_sqs import SQSClient
    from typing_extensions import Literal


# SEE ALSO ec2_alloc_stub.py

_T = TypeVar("_T")
_U = TypeVar("_U")

# AMIs that have meadowrun pre-installed. These are all identical, we just need to
# replicate into each region.
_AMIS = {
    "plain": {
        "us-east-2": "ami-01fca960c9299b335",
        "us-east-1": "ami-03605d9958c76f9ed",
        "us-west-1": "ami-0f4875d5d9c68dff4",
        "us-west-2": "ami-0e6855d6badbfd1bb",
        "eu-central-1": "ami-00d7b0f1c05591b34",
        "eu-west-1": "ami-0e6989e6e25e17893",
        "eu-west-2": "ami-008e7c6aecbbcd3e9",
        "eu-west-3": "ami-0b2ab4237804283b3",
        "eu-north-1": "ami-0d41c6161b3bdf6b9",
    },
    "cuda": {
        "us-east-2": "ami-0da59419ffa313be5",
        "us-east-1": "ami-0c2990d11681b3c3c",
        "us-west-1": "ami-0e7675039071f6674",
        "us-west-2": "ami-0811b0dd8b334d076",
        "eu-central-1": "ami-0706e0cc9199ea6a3",
        "eu-west-1": "ami-0d939ea5b21e04a2c",
        "eu-west-2": "ami-0bd66e6a4b8dbd3e2",
        "eu-west-3": "ami-0ac87c5282b090f4e",
        "eu-north-1": "ami-0f663e908bc88c42c",
    },
}
SSH_USER = "ubuntu"


def ensure_ec2_alloc_table(region_name: str) -> None:
    """Creates the EC2 alloc DynamoDB table if it doesn't already exist"""
    db = boto3.resource("dynamodb", region_name=region_name)

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/dynamodb.html#DynamoDB.ServiceResource.create_table
    ignore_boto3_error_code(
        lambda: db.create_table(
            TableName=_EC2_ALLOC_TABLE_NAME,
            AttributeDefinitions=[
                {
                    "AttributeName": _INSTANCE_ID,
                    "AttributeType": "S",
                }
            ],
            KeySchema=[
                {
                    "AttributeName": _INSTANCE_ID,
                    "KeyType": "HASH",
                }
            ],
            BillingMode="PAY_PER_REQUEST",
            TableClass="STANDARD",
        ),
        "ResourceInUseException",
    )

    db.Table(_EC2_ALLOC_TABLE_NAME).wait_until_exists()


class EC2InstanceRegistrar(InstanceRegistrar[_InstanceState]):
    """
    The EC2 instance registrar uses a DynamoDB table to keep track of instances and job
    allocations.
    """

    def __init__(
        self,
        table_region_name: Optional[str],
        on_table_missing: Literal["create", "raise"],
    ):
        self._table_region_name = table_region_name
        self._on_table_missing = on_table_missing
        self._table: Any = None

    async def __aenter__(self) -> EC2InstanceRegistrar:
        """
        Gets the EC2 instances table if it exists, otherwise creates it and then gets
        the newly created table.

        See register_instance for a description of the expected schema
        """

        if self._table_region_name is None:
            self._table_region_name = await _get_default_region_name()

        db = boto3.resource("dynamodb", region_name=self._table_region_name)
        self._table = db.Table(_EC2_ALLOC_TABLE_NAME)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def get_region_name(self) -> str:
        """Just for testing"""
        if self._table_region_name is None:
            raise ValueError(
                "Programming error: requested table_region_name but __aenter__ was not "
                "called to initialize"
            )
        return self._table_region_name

    async def register_instance(
        self,
        public_address: str,
        name: str,
        resources_available: ResourcesInternal,
        running_jobs: List[Tuple[str, ResourcesInternal]],
    ) -> None:
        # TODO we should probably enforce types on logical_cpu_available and
        # memory_gb_available

        now = datetime.datetime.utcnow().isoformat()

        success, result = ignore_boto3_error_code(
            lambda: self._table.put_item(
                Item={
                    # the public address of the EC2 instance
                    _PUBLIC_ADDRESS: public_address,
                    _INSTANCE_ID: name,
                    # the resources (e.g. CPU/memory) available on the EC2 instance
                    # (after allocating resources to the already running jobs below)
                    _RESOURCES_AVAILABLE: resources_available.consumable_as_decimals(),
                    _NON_CONSUMABLE_RESOURCES: (
                        resources_available.non_consumable_as_decimals()
                    ),
                    # the jobs currently allocated to run on this instance
                    _RUNNING_JOBS: {
                        job_id: {
                            # Represents how many resources this job is "using". These
                            # values will be used to add resources back to the
                            # _AVAILABLE* fields when the jobs get deallocated
                            _RESOURCES_ALLOCATED: (
                                allocated_resources.consumable_as_decimals()
                            ),
                            # When the job was allocated. This gets used by
                            # deallocate_tasks.py in the case where a client allocates a
                            # job but crashes before it can launch the job. After a
                            # timeout, deallocate_tasks.py will assume this job will
                            # never start and deallocates it.
                            _ALLOCATED_TIME: now,
                        }
                        for job_id, allocated_resources in running_jobs
                    },
                    # The last time a job was allocated or deallocated to this machine
                    _LAST_UPDATE_TIME: now,
                    # Normally False. Set to True for cases like when a spot instance is
                    # being interrupted so we shouldn't assign any new jobs to this
                    # instance
                    _PREVENT_FURTHER_ALLOCATION: False,
                },
                ConditionExpression=f"attribute_not_exists({_PUBLIC_ADDRESS})",
            ),
            "ConditionalCheckFailedException",
        )

        if not success:
            # It's possible that an existing EC2 instance crashed unexpectedly, the
            # coordinator record hasn't been deleted yet, and a new instance was created
            # that has the same address
            raise ValueError(
                f"Tried to register an ec2_instance {public_address} but it already "
                f"exists, this should never happen!"
            )

    async def get_registered_instances(self) -> List[_InstanceState]:
        response = self._table.scan(
            Select="SPECIFIC_ATTRIBUTES",
            ProjectionExpression=",".join(
                [
                    _PUBLIC_ADDRESS,
                    _INSTANCE_ID,
                    _RESOURCES_AVAILABLE,
                    _NON_CONSUMABLE_RESOURCES,
                    _PREVENT_FURTHER_ALLOCATION,
                ]
            ),
        )

        if response.get("LastEvaluatedKey"):
            # TODO scan maxes out at 1MB of returned data before requiring pagination.
            # We could implement pagination, but scanning for so much data will cost a
            # lot of money, so we should eventually add either secondary indexes that
            # let us hone in on the instances we want (e.g. exclude instances that don't
            # have enough resources to run our job) or limit the scan to some number of
            # instances and not do a global optimization
            raise NotImplementedError(
                "We don't currently support a very large number of EC2 instances"
            )

        return [
            _InstanceState(
                item[_PUBLIC_ADDRESS],
                item[_INSTANCE_ID],
                ResourcesInternal.from_decimals(
                    item[_RESOURCES_AVAILABLE], item[_NON_CONSUMABLE_RESOURCES]
                ),
                # This is a bit of hack, but for the EC2InstanceRegistrar, we know that
                # we won't need the running_jobs in the context that this function is
                # called, so we just set it to None to save bandwidth/memory/etc.
                None,
                item[_PREVENT_FURTHER_ALLOCATION],
            )
            for item in response["Items"]
        ]

    async def get_registered_instance(self, name: str) -> _InstanceState:
        result = self._table.get_item(
            Key={_INSTANCE_ID: name},
            ProjectionExpression=",".join(
                [_RUNNING_JOBS, _PUBLIC_ADDRESS, _PREVENT_FURTHER_ALLOCATION]
            ),
        )
        if "Item" not in result:
            raise ValueError(f"ec2 instance {name} was not found")

        return _InstanceState(
            result["Item"][_PUBLIC_ADDRESS],
            name,
            # similar to above, also a hack. We know we won't need the
            # available_resources in the context that this function is called
            None,
            result["Item"][_RUNNING_JOBS],
            result["Item"][_PREVENT_FURTHER_ALLOCATION],
        )

    async def allocate_jobs_to_instance(
        self,
        instance: _InstanceState,
        resources_allocated_per_job: ResourcesInternal,
        new_job_ids: List[str],
    ) -> bool:
        if len(new_job_ids) == 0:
            raise ValueError("Must provide at least one new_job_ids")

        now = datetime.datetime.utcnow().isoformat()

        expression_attribute_names = {}
        set_expressions = [f"{_LAST_UPDATE_TIME}=:now"]
        expression_attribute_values: Dict[str, object] = {":now": now}
        conditional_expressions = []

        for i, (key, value) in enumerate(
            resources_allocated_per_job.multiply(len(new_job_ids)).consumable.items()
        ):
            # resource names might not be valid dynamodb identifiers
            expression_attribute_names[f"#r{i}"] = key
            # subtract the resources used
            set_expressions.append(
                f"{_RESOURCES_AVAILABLE}.#r{i} = {_RESOURCES_AVAILABLE}.#r{i} - :r{i}"
            )
            expression_attribute_values[f":r{i}"] = decimal.Decimal(value)
            # check that we don't use resources we don't have
            conditional_expressions.append(f"{_RESOURCES_AVAILABLE}.#r{i} >= :r{i}")
            # non_consumable resources need to be checked by the caller

        for i, job_id in enumerate(new_job_ids):
            # job_ids might not be valid dynamodb identifiers
            expression_attribute_names[f"#j{i}"] = job_id
            # add the jobs with their metadata to _RUNNING_JOBS
            set_expressions.append(f"{_RUNNING_JOBS}.#j{i} = :j{i}")
            expression_attribute_values[f":j{i}"] = {
                _RESOURCES_ALLOCATED: (
                    resources_allocated_per_job.consumable_as_decimals()
                ),
                _ALLOCATED_TIME: now,
            }
            # check that the job_id doesn't already exist
            conditional_expressions.append(
                f"attribute_not_exists({_RUNNING_JOBS}.#j{i})"
            )

        success, result = ignore_boto3_error_code(
            lambda: self._table.update_item(
                Key={_INSTANCE_ID: instance.name},
                # subtract resources that we're allocating
                UpdateExpression="SET " + ", ".join(set_expressions),
                # Check to make sure the allocation is still valid
                ConditionExpression=" AND ".join(conditional_expressions),
                ExpressionAttributeValues=expression_attribute_values,
                ExpressionAttributeNames=expression_attribute_names,
            ),
            "ConditionalCheckFailedException",
        )
        return success

    def allocate_jobs_to_instance_max_chunk(self) -> int:
        # the limit for update/conditional expressions in DynamoDB is 4KB:
        # https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/ServiceQuotas.html
        # We hit this at around 90 jobs, and we give ourselves a bit of extra buffer
        # here
        return 85

    async def deallocate_job_from_instance(
        self, instance: _InstanceState, job_id: str
    ) -> bool:
        job = instance.get_running_jobs()[job_id]

        expression_attribute_names = {"#job_id": job_id}
        set_expressions = [f"{_LAST_UPDATE_TIME}=:now"]
        expression_attribute_values: Dict[str, object] = {
            ":now": datetime.datetime.utcnow().isoformat()
        }

        for i, (key, value) in enumerate(job[_RESOURCES_ALLOCATED].items()):
            # resource names might not be valid dynamodb identifiers
            expression_attribute_names[f"#r{i}"] = key
            # add back the resources that were held by the job
            set_expressions.append(
                f"{_RESOURCES_AVAILABLE}.#r{i} = {_RESOURCES_AVAILABLE}.#r{i} + :r{i}"
            )
            expression_attribute_values[f":r{i}"] = decimal.Decimal(value)

        success, result = ignore_boto3_error_code(
            lambda: self._table.update_item(
                Key={_INSTANCE_ID: instance.name},
                UpdateExpression=(
                    f"SET {', '.join(set_expressions)} REMOVE {_RUNNING_JOBS}.#job_id"
                ),
                # make sure we haven't already removed this job
                ConditionExpression=f"attribute_exists({_RUNNING_JOBS}.#job_id)",
                ExpressionAttributeNames=expression_attribute_names,
                ExpressionAttributeValues=expression_attribute_values,
            ),
            "ConditionalCheckFailedException",
        )
        return success

    async def set_prevent_further_allocation(self, name: str, value: bool) -> bool:
        self._table.update_item(
            Key={_INSTANCE_ID: name},
            UpdateExpression=(
                f"SET {_PREVENT_FURTHER_ALLOCATION}=:value, {_LAST_UPDATE_TIME}=:now"
            ),
            ExpressionAttributeValues={
                ":value": value,
                ":now": datetime.datetime.utcnow().isoformat(),
            },
        )
        return True

    async def launch_instances(
        self,
        instance_type_resources_required_per_task: ResourcesInternal,
        num_concurrent_tasks: int,
        alloc_cloud_instances: AllocVM,
        abort: Optional[asyncio.Event],
    ) -> Sequence[CloudInstance]:
        if not isinstance(alloc_cloud_instances, AllocEC2Instance):
            # TODO do this in the type checker somehow
            raise ValueError(
                "Programming error: EC2InstanceRegistrar can only be used with "
                "AllocEC2Instance"
            )

        region_name = alloc_cloud_instances._get_region_name()

        if alloc_cloud_instances.ami_id:
            ami = alloc_cloud_instances.ami_id
        else:
            if "nvidia" in instance_type_resources_required_per_task.non_consumable:
                ami_type = "cuda"
            else:
                ami_type = "plain"

            if region_name not in _AMIS[ami_type]:
                raise ValueError(
                    f"The meadowrun AMI is not available in {region_name}. Please ask "
                    f"the meadowrun maintainers to add support for this region: "
                    "https://github.com/meadowdata/meadowrun/issues and try a supported"
                    f" region for now: {', '.join(_AMIS[ami_type].keys())}"
                )
            ami = _AMIS[ami_type][region_name]

        if alloc_cloud_instances.subnet_id:
            subnet_id = alloc_cloud_instances.subnet_id
        else:
            subnet_id = None

        if alloc_cloud_instances.security_group_ids:
            if isinstance(alloc_cloud_instances.security_group_ids, str):
                security_group_ids: Sequence[str] = [
                    alloc_cloud_instances.security_group_ids
                ]
            else:
                security_group_ids = alloc_cloud_instances.security_group_ids
        else:
            if subnet_id is not None:
                subnet = boto3.client("ec2", region_name=region_name).describe_subnets(
                    SubnetIds=[subnet_id]
                )["Subnets"][0]
                vpc_id = subnet["VpcId"]
            else:
                vpc_id = None

            security_group_ids = [get_ssh_security_group_id(region_name, vpc_id)]

        if alloc_cloud_instances.iam_role_instance_profile:
            iam_role_instance_profile = alloc_cloud_instances.iam_role_instance_profile
        else:
            iam_role_instance_profile = _EC2_ROLE_INSTANCE_PROFILE

        return await launch_ec2_instances(
            instance_type_resources_required_per_task,
            num_concurrent_tasks,
            LaunchEC2InstanceSettings(
                ami_id=ami,
                subnet_id=subnet_id,
                security_group_ids=security_group_ids,
                iam_role_instance_profile=iam_role_instance_profile,
            ),
            region_name=region_name,
            # assumes that we've already called ensure_meadowrun_key_pair!
            key_name=MEADOWRUN_KEY_PAIR_NAME,
            tags={_EC2_ALLOC_TAG: _EC2_ALLOC_TAG_VALUE},
            abort=abort,
        )

    async def authorize_current_ip(self, alloc_cloud_instances: AllocVM) -> None:
        if not isinstance(alloc_cloud_instances, AllocEC2Instance):
            # TODO do this in the type checker somehow
            raise ValueError(
                "Programming error: EC2InstanceRegistrar can only be used with "
                "AllocEC2Instance"
            )

        await authorize_current_ip_for_meadowrun_ssh(
            self.get_region_name(), alloc_cloud_instances.subnet_id
        )

    async def open_ports(
        self,
        ports: Optional[Sequence[str]],
        allocated_existing_instances: Iterable[_TInstanceState],
        allocated_new_instances: Iterable[CloudInstance],
    ) -> None:
        if ports:
            security_group_ids = await asyncio.gather(
                *(ensure_port_security_group(p, self.get_region_name()) for p in ports)
            )
            ec2 = boto3.client("ec2", region_name=self.get_region_name())
            for instance_id in itertools.chain(
                (i.name for i in allocated_existing_instances),
                (i.name for i in allocated_new_instances),
            ):
                groups = {
                    group["GroupId"]
                    for group in ec2.describe_instance_attribute(
                        InstanceId=instance_id, Attribute="groupSet"
                    )["Groups"]
                }
                groups.update(security_group_ids)
                ec2.modify_instance_attribute(
                    InstanceId=instance_id, Groups=list(groups)
                )


@dataclasses.dataclass()
class AllocEC2Instance(AllocVM):
    """
    Specifies that the job should be run on a dynamically allocated EC2 instance. Any
    existing Meadowrun-managed EC2 instances will be reused if available. If none are
    available, Meadowrun will launch the cheapest instance type that meets the resource
    requirements for a job.

    `resources_required` must be provided with the AllocEC2Instance Host.

    Attributes:
        region_name: Specifies the region name for EC2, e.g. "us-east-2". None will use
            the default region_name.
        ami_id: An AMI ID for EC2, e.g. "ami-006426834f282c3d7". This image must be
            available in the specified region. The AMI specified must be built off of
            the Meadowrun AMIs as Meadowrun expects certain python environments and
            folders to be available. See [Use a custom AMI (machine image) on
            AWS](../../how_to/custom_ami)

            If this is not specified, Meadowrun will use the default Meadowrun-supplied
            images when launching new images. If there are existing instances, the job
            will run on any instance, regardless of what image was used to launch it.
        subnet_id: The subnet id for EC2, e.g. "subnet-02e4e2996d2fb96d9". The subnet
            must be in the specified region. If this is not specified, the default
            subnet in the default VPC will be used.

            If you specify a subnet, you'll need to think about how to connect to the
            machine in your subnet. One option is to [set your subnet to auto-assign IP
            addresses](https://docs.aws.amazon.com/vpc/latest/userguide/configure-subnets.html#subnet-settings).
            If an instance does not have a public IP address, Meadowrun will try to
            connect via the private IP address. This will only work if the machine
            launching the job can access the private IP address. The default subnet in a
            new AWS account will be set to auto-assign IP addresses.

            If your subnet does not have access to the internet, you'll need to make
            sure that any dependencies (e.g. pip packages, container images) that
            Meadowrun will try to download are available without internet access. The
            default subnet in a new AWS account will have the appropriate Route Tables
            and Internet Gateway configuration to be able to access the internet.
        security_group_ids: A list of security group ids, e.g. "sg-0690853a8374b9b6d".
            The security group must be in the same VPC as the specified subnet_id
            (security groups are specific to each VPC), and it must allow you to SSH
            over port 22 to the machine from your current IP address. If this is not
            specified, Meadowrun will use a security group created at install time in
            the default VPC. Meadowrun's install step sets up a security group and opens
            port 22 for that security group for the current IP whenever Meadowrun is
            used. If you specify a subnet that is not in the default VPC, this parameter
            is required, as the default security group will not be available. Please
            also see the "ports" argument on the run_* commands.
        iam_role_instance_profile: The name of an instance profile for an IAM role name
            (not to be confused with the IAM role itself!), e.g.
            "meadowrun_ec2_role_instance_profile". The EC2 instance will be launched
            under this IAM role. By default, Meadowrun will use an IAM role created at
            install time called meadowrun_ec2_role that has the permissions needed for a
            Meadowrun-managed EC2 instance. Any IAM role you specify must have a
            superset of the permissions granted by meadowrun_ec2_role. The easiest way
            to implement this is to attach the Meadowrun-generated
            "meadowrun_ec2_policy" to your IAM role (in addition to any custom policies
            you wish to add). Please also see [Access resources from Meadowrun
            jobs](../../how_to/access_resources)
    """

    region_name: Optional[str] = None
    ami_id: Union[str, None] = None
    subnet_id: Optional[str] = None
    security_group_ids: Union[str, Sequence[str], None] = None
    iam_role_instance_profile: Optional[str] = None

    def get_cloud_provider(self) -> CloudProviderType:
        return "EC2"

    async def set_defaults(self) -> None:
        # This does not need to set defaults for ami_id, subnet_id, etc., as those can
        # all be set in run_job. The only thing we need to set at this point is the
        # region_name, because an S3 code upload requires the region to be decided.
        if not self.region_name:
            self.region_name = await _get_default_region_name()

    def _get_region_name(self) -> str:
        # Wrapper around region_name that throws if it is None. Should only be used
        # internally.
        if self.region_name is None:
            raise ValueError(
                "Programming error: region_name is None but it should have been set by "
                "set_defaults earlier"
            )
        return self.region_name

    def get_runtime_resources(self) -> ResourcesInternal:
        non_consumable = {}
        if self.ami_id is not None:
            non_consumable[self.ami_id] = 1.0
        if self.subnet_id is not None:
            non_consumable[self.subnet_id] = 1.0
        if self.security_group_ids:
            if isinstance(self.security_group_ids, str):
                non_consumable[self.security_group_ids] = 1.0
            else:
                for security_group_id in self.security_group_ids:
                    non_consumable[security_group_id] = 1.0
        if self.iam_role_instance_profile:
            non_consumable[self.iam_role_instance_profile] = 1.0

        return ResourcesInternal({}, non_consumable)

    async def run_job(
        self,
        resources_required: Optional[ResourcesInternal],
        job: Job,
        wait_for_result: WaitOption,
    ) -> JobCompletion[Any]:
        if resources_required is None:
            raise ValueError(
                "Resources.logical_cpu and memory_gb must be specified for "
                "AllocEC2Instance"
            )

        return await run_job_ec2_instance_registrar(
            job, resources_required, self, wait_for_result
        )

    def _create_grid_job_cloud_interface(self) -> GridJobCloudInterface:
        return EC2GridJobInterface(self)

    async def get_storage_bucket(self) -> AbstractStorageBucket:
        return get_aws_s3_bucket(self._get_region_name())


async def run_job_ec2_instance_registrar(
    job: Job,
    resources_required: ResourcesInternal,
    alloc_ec2_instance: AllocEC2Instance,
    wait_for_result: WaitOption,
) -> JobCompletion[Any]:
    """Runs the specified job on EC2. Creates an EC2InstanceRegistrar"""
    region_name = alloc_ec2_instance._get_region_name()
    pkey = get_meadowrun_ssh_key(region_name)

    async with EC2InstanceRegistrar(region_name, "create") as instance_registrar:
        host, job_id = await allocate_single_job_to_instance(
            instance_registrar,
            resources_required,
            alloc_ec2_instance,
            job.ports,
        )

    # Kind of weird that we're changing the job_id here, but okay as long as job_id
    # remains mostly an internal concept
    job.job_id = job_id

    return await SshHost(host, SSH_USER, pkey, ("EC2", region_name)).run_job(
        resources_required, job, wait_for_result
    )


class EC2GridJobInterface(GridJobCloudInterface):
    """
    This is a relatively thin wrapper around some functionality in grid_tasks_sqs. This
    class should be in grid_tasks_sqs, but it's here because of circular import issues.
    """

    def __init__(self, alloc_cloud_instance: AllocEC2Instance):
        self._cloud_provider = alloc_cloud_instance.get_cloud_provider()
        self._region_name = alloc_cloud_instance._get_region_name()

        self._ssh_private_key = get_meadowrun_ssh_key(self._region_name)

        # We keep multiple request queues so that we can retry tasks with more
        # resources. The ith queue (where i starts at 0) will have workers with
        # (original_memory * (i + 1)) memory. For the happy path, we will only create
        # one queue with one set of workers having the originally requested memory
        self._request_queue_urls: List[asyncio.Task[str]] = []
        self._task_argument_ranges: Optional[asyncio.Task[List[Tuple[int, int]]]] = None

        # this id is just used for creating the job's queues. It has no relationship to
        # any Job.job_ids
        self._job_id = str(uuid.uuid4())

        self._sqs_client: Optional[SQSClient] = None
        self._s3_bucket: Optional[S3Bucket] = None

    async def __aenter__(self) -> EC2GridJobInterface:
        session = aiobotocore.session.get_session()
        self._sqs_client = await session.create_client(
            "sqs", region_name=self._region_name
        ).__aenter__()
        self._s3_bucket = await get_aws_s3_bucket(self._region_name).__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        if self._sqs_client is not None:
            await self._sqs_client.__aexit__(exc_type, exc_val, exc_tb)
        if self._s3_bucket is not None:
            await self._s3_bucket.__aexit__(exc_type, exc_val, exc_tb)

    def create_instance_registrar(self) -> InstanceRegistrar:
        return EC2InstanceRegistrar(self._region_name, "create")

    def create_queue(self) -> int:
        if self._sqs_client is None:
            raise ValueError("EC2GridJobInterface must be created with `async with`")

        index = len(self._request_queue_urls)
        if index == 0:
            job_id_modified = self._job_id
        else:
            job_id_modified = f"{self._job_id}-{index}"
        self._request_queue_urls.append(
            asyncio.create_task(create_request_queue(job_id_modified, self._sqs_client))
        )
        return index

    async def setup_and_add_tasks(self, tasks: Sequence[_T]) -> None:
        if self._sqs_client is None or self._s3_bucket is None:
            raise ValueError("EC2GridJobInterface must be created with `async with`")

        # create SQS queues and add tasks to the request queue
        print(f"The current run_map's id is {self._job_id}")
        queue_index = self.create_queue()
        if queue_index != 0:
            raise ValueError(
                "setup_and_add_tasks was called more than once, or create_queue was "
                "called before setup_and_add_tasks"
            )
        self._task_argument_ranges = asyncio.create_task(
            add_tasks(
                self._job_id,
                await self._request_queue_urls[queue_index],
                self._s3_bucket,
                self._sqs_client,
                tasks,
            )
        )
        await self._task_argument_ranges

    async def ssh_host_from_address(self, address: str, instance_name: str) -> SshHost:
        return SshHost(
            address,
            SSH_USER,
            self._ssh_private_key,
            (self._cloud_provider, self._region_name),
            instance_name,
        )

    async def shutdown_workers(self, num_workers: int, queue_index: int) -> None:
        if len(self._request_queue_urls) < queue_index + 1:
            raise ValueError(f"Queue {queue_index} has not been created yet")
        if self._sqs_client is None:
            raise ValueError("EC2GridJobInterface must be created with `async with`")
        await add_worker_shutdown_messages(
            await self._request_queue_urls[queue_index], num_workers, self._sqs_client
        )

    async def get_worker_function(
        self, queue_index: int
    ) -> Callable[
        [str, str, TaskWorkerServer, WorkerMonitor],
        Coroutine[Any, Any, None],
    ]:
        if len(self._request_queue_urls) < queue_index + 1:
            raise ValueError(f"Queue {queue_index} has not been created yet")

        return functools.partial(
            worker_function,
            await self._request_queue_urls[queue_index],
            self._job_id,
            self._region_name,
        )

    async def receive_task_results(
        self, *, stop_receiving: asyncio.Event, workers_done: asyncio.Event
    ) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:
        if self._s3_bucket is None:
            raise ValueError("EC2GridJobInterface must be created with `async with`")

        # Note: download here is all via S3. It's important that downloads are fast
        # enough - in some cases (many workers, small-ish tasks) the rate of downloading
        # the results to the client can be a limiting factor. Other alternatives that
        # were considered: - SQS queues. Limitations are 1. Only 10 messages can be
        # downloaded at a time, which slows us down when we have hundreds of tasks
        # completing quickly. 2. SQS message limit is 256KB which means we need to use
        # S3 to transfer the actual data in the case of large results. - Sending data
        # via SSH/SCP. This seems like it should be faster especially in the same
        # VPC/subnet, but even in that case, uploading to S3 first seems to be faster.
        # It's possible this would make sense in cases where we e.g. overwhelm the S3
        # bucket.

        return receive_results(
            self._s3_bucket,
            self._job_id,
            stop_receiving=stop_receiving,
            all_workers_exited=workers_done,
            # it's rare for any results to be ready in <4 seconds
            initial_wait_seconds=4,
        )

    async def retry_task(
        self, task_id: int, attempts_so_far: int, queue_index: int
    ) -> None:
        if len(self._request_queue_urls) < queue_index + 1:
            raise ValueError(f"Queue {queue_index} has not been created yet")
        if self._task_argument_ranges is None:
            raise ValueError("setup_and_add_tasks must be called before retry_task")
        if self._sqs_client is None:
            raise ValueError("EC2GridJobInterface must be created with `async with`")

        await retry_task(
            await self._request_queue_urls[queue_index],
            task_id,
            attempts_so_far + 1,
            (await self._task_argument_ranges)[task_id],
            self._sqs_client,
        )



"""
Credentials can be managed in a few different ways. Let's say we have a
username/password that we need to access a Docker container registry. The simplest thing
to do is for the user to send the username/password to the coordinator, which then sends
it to agents when the coordinator assigns them a job that needs it. Alternatively, we
could just send over the secret if the agent actually needs it, i.e. the image is not in
the agent machine's local cache.

It's probably better for the user to send over a credentials source, e.g. an AWS secret
or even just a file path that is restricted in some way. If the user is accessing the
coordinator over the public internet, this makes it possible to avoid sending the
username/password over the public internet (even if it's encrypted). Also, this makes it
possible to rotate secrets without manually updating the coordinator. If we take this
strategy, then we need to decide whether the coordinator should send over the actual
credentials or the credentials source to the agents. Sending the actual credentials
exposes them over the wire (even if they're encrypted), but sending the credentials
source means that all of the agents need direct access to the secrets. If e.g. an AWS
IAM role is used to restrict access to an AWS secret, then this is not good because the
job then gets full access to that AWS secret.

TODO actually encrypt the credentials as we send them over?
"""
from __future__ import annotations

import abc
import dataclasses
import json
from typing import Union, Optional, Dict, List, Tuple

import boto3
from typing_extensions import Literal

import meadowrun.docker_controller
from meadowrun.aws_integration.aws_core import _get_default_region_name
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
)
from meadowrun.meadowrun_pb2 import (
    AwsSecretProto,
    AzureSecretProto,
    Credentials,
    KubernetesSecretProto,
    ServerAvailableFile,
)

# Represents a way to get credentials
CredentialsSource = Union[
    AwsSecretProto, AzureSecretProto, KubernetesSecretProto, ServerAvailableFile
]
# Represents a service that credentials can be used for
CredentialsService = Literal["DOCKER", "GIT"]


@dataclasses.dataclass(frozen=True)
class CredentialsSourceForService:
    """A CredentialsSource with metadata about what service it should be used for"""

    service: CredentialsService
    service_url: str
    source: CredentialsSource


# Maps from a Credentials.Service to a list of (service_url, CredentialsSource). This is
# usually how we'll store a set of CredentialsSources
CredentialsDict = Dict[
    "Credentials.Service.ValueType", List[Tuple[str, CredentialsSource]]
]


class RawCredentials(abc.ABC):
    """
    Represents credentials that can be used immediately (i.e. as opposed to a
    CredentialsSource)
    """

    pass


@dataclasses.dataclass()
class UsernamePassword(RawCredentials):
    username: str
    password: str


@dataclasses.dataclass()
class SshKey(RawCredentials):
    """Should be the content of the private key file generated by e.g. ssh-keygen"""

    private_key: str


@dataclasses.dataclass(frozen=True)
class KubernetesSecretRaw(RawCredentials):
    """
    This class is a bit silly, as we're kind of just maintaining a fiction that
    Kubernetes Job objects need to be serializable as protobufs. If we wanted to break
    that self-imposed convention, we could just send a direct KubernetesSecret object
    through the Job.

    We generally don't want to "unpack" Kubernetes secrets ourselves, as Kubernetes will
    do that for us.
    """

    secret_name: str


async def _get_credentials_from_source(source: CredentialsSource) -> RawCredentials:
    if isinstance(source, AwsSecretProto):
        # TODO not sure if it's better to try to reuse a client/session or just create a
        #  new one each time? This seems related:
        #  https://github.com/boto/botocore/issues/619
        secret = json.loads(
            boto3.client(
                service_name="secretsmanager",
                region_name=await _get_default_region_name(),
            ).get_secret_value(SecretId=source.secret_name)["SecretString"]
        )
        if source.credentials_type == Credentials.Type.USERNAME_PASSWORD:
            return UsernamePassword(secret["username"], secret["password"])
        elif source.credentials_type == Credentials.Type.SSH_KEY:
            return SshKey(secret["private_key"])
        else:
            raise ValueError(f"Unknown credentials type {source.credentials_type}")
    elif isinstance(source, AzureSecretProto):
        result = await azure_rest_api(
            "GET",
            f"secrets/{source.secret_name}",
            "7.3",
            base_url=source.vault_name,
            token_scope="https://vault.azure.net/.default",
        )
        secret_value = result["value"]
        if source.credentials_type == Credentials.Type.USERNAME_PASSWORD:
            secret_json = json.loads(secret_value)
            return UsernamePassword(secret_json["username"], secret_json["password"])
        elif source.credentials_type == Credentials.Type.SSH_KEY:
            return SshKey(secret_value)
        else:
            raise ValueError(f"Unknown credentials type {source.credentials_type}")
    elif isinstance(source, ServerAvailableFile):
        if source.credentials_type == Credentials.Type.USERNAME_PASSWORD:
            with open(source.path, "r", encoding="utf-8") as f:
                lines = f.readlines()
            if len(lines) != 2:
                raise ValueError(
                    "ServerAvailableFile for credentials must have exactly two lines, "
                    "one for username, and one for password"
                )
            # strip just the trailing newlines, other whitespace might be needed
            lines = [line.rstrip("\r\n") for line in lines]

            return UsernamePassword(lines[0], lines[1])
        elif source.credentials_type == Credentials.Type.SSH_KEY:
            with open(source.path, "r", encoding="utf-8") as f:
                return SshKey(f.read())
        else:
            raise ValueError(f"Unknown credentials type {source.credentials_type}")
    elif isinstance(source, KubernetesSecretProto):
        return KubernetesSecretRaw(source.secret_name)
    else:
        raise ValueError(f"Unknown type of credentials source {type(source)}")


async def get_matching_credentials(
    service: Credentials.Service.ValueType,
    service_url_to_match: str,
    credentials_dict: CredentialsDict,
) -> Optional[RawCredentials]:
    """
    Gets the credentials where the service_url matches our normalized_url, and then pick
    the longest one (if it exists). max picks the first element it sees, and credentials
    are in the order they were added, so we reverse the list of credentials first so
    that we get the latest set of credentials if multiple credentials are the longest
    """
    # TODO we should clean up "replaced" credentials that will never get used again
    source = max(
        (
            (service_url, source)
            for service_url, source in reversed(credentials_dict.get(service, ()))
            if service_url_to_match.startswith(service_url)
        ),
        key=lambda c: len(c[0]),
        default=None,
    )
    if source is not None:
        return await _get_credentials_from_source(source[1])
    else:
        return None


async def get_docker_credentials(
    repository: str, credentials_dict: CredentialsDict
) -> Optional[RawCredentials]:
    """Get credentials for a docker repository"""

    registry_domain, repository_name = meadowrun.docker_controller.get_registry_domain(
        repository
    )
    return await get_matching_credentials(
        Credentials.Service.DOCKER,
        f"{registry_domain}/{repository_name}",
        credentials_dict,
    )



from __future__ import annotations

from typing import List, Tuple, Optional, Type, Coroutine, Any, TYPE_CHECKING

import aiohttp
import meadowrun._vendor.gcloud.aio.storage as gcloud_storage
from meadowrun.storage_grid_job import AbstractStorageBucket

if TYPE_CHECKING:
    from types import TracebackType


class GoogleStorageBucket(AbstractStorageBucket):
    def __init__(self, storage: gcloud_storage.Storage, bucket: str):
        self._storage = storage
        self._bucket = bucket

    def get_cache_key(self) -> str:
        return f"gcp/{self._bucket}"

    async def __aenter__(self) -> GoogleStorageBucket:
        self._storage = await self._storage.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self._storage.__aexit__(exc_type, exc_val, exc_tb)

    def get_bytes(self, key: str) -> Coroutine[Any, Any, bytes]:
        return self._storage.download(self._bucket, key)

    async def try_get_bytes(self, key: str) -> Optional[bytes]:
        try:
            return await self.get_bytes(key)
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return None
            raise

    def get_byte_range(
        self, key: str, byte_range: Tuple[int, int]
    ) -> Coroutine[Any, Any, bytes]:
        return self._storage.download(
            self._bucket,
            key,
            headers={"Range": f"bytes={byte_range[0]}-{byte_range[1]}"},
        )

    async def write_bytes(self, data: bytes, key: str) -> None:
        await self._storage.upload(self._bucket, key, data)

    async def exists(self, key: str) -> bool:
        try:
            await self._storage.download_metadata(self._bucket, key)
            return True
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return False
            raise

    def get_file(self, key: str, local_filename: str) -> Coroutine[Any, Any, None]:
        return self._storage.download_to_filename(self._bucket, key, local_filename)

    async def try_get_file(self, key: str, local_filename: str) -> bool:
        try:
            await self.get_file(key, local_filename)
            return True
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return False
            raise

    async def write_file(self, local_filename: str, key: str) -> None:
        await self._storage.upload_from_filename(self._bucket, key, local_filename)

    async def list_objects(self, key_prefix: str) -> List[str]:
        result = await self._storage.list_objects(
            self._bucket, params={"prefix": key_prefix, "fields": "items(name)"}
        )
        return [item["name"] for item in result.get("items", ())]

    async def delete_object(self, key: str) -> None:
        await self._storage.delete(self._bucket, key)


def get_google_storage_bucket(bucket: str) -> GoogleStorageBucket:
    return GoogleStorageBucket(gcloud_storage.Storage(), bucket)



from __future__ import annotations

import xml.dom.minidom
from typing import TYPE_CHECKING, Any, Iterable, Optional, Tuple, Type

if TYPE_CHECKING:
    import aiohttp
    import requests


class AzureRestApiError(Exception):
    """
    status is the integer http status code. code and message are typically returned by
    Azure APIs. code is usually a single word/phrase like "ResourceNotFound", and
    message is usually a more verbose explanation.
    """

    def __init__(self, status: int, code: Optional[str], message: str):
        self.status = status
        self.code = code
        self.message = message


class ResourceNotFoundError(AzureRestApiError):
    def __init__(self, status: int, code: Optional[str], message: str):
        super().__init__(status, code, message)


class ResourceExistsError(AzureRestApiError):
    def __init__(self, status: int, code: Optional[str], message: str):
        super().__init__(status, code, message)


class ResourceModifiedError(AzureRestApiError):
    def __init__(self, status: int, code: Optional[str], message: str):
        super().__init__(status, code, message)


def _get_code_and_message_from_json(
    response_json: Any,
) -> Tuple[Optional[str], Optional[str]]:
    if "error" in response_json:
        error = response_json["error"]
    elif "errors" in response_json and len(response_json["errors"]) == 1:
        error = response_json["errors"][0].get("code")
    elif "odata.error" in response_json:
        error = response_json["odata.error"]
    else:
        error = None

    if error is not None:
        return error.get("code"), error.get("message")
    else:
        return None, None


def _get_code_and_message_from_xml(
    response_text: str,
) -> Tuple[Optional[str], Optional[str]]:
    code, message = None, None
    try:
        parsed = xml.dom.minidom.parseString(response_text)
        if parsed and len(parsed.documentElement.childNodes) == 2:
            if parsed.documentElement.childNodes[0].nodeName == "Code":
                code = parsed.documentElement.childNodes[0].childNodes[0].nodeValue
            if parsed.documentElement.childNodes[1].nodeName == "Message":
                message = parsed.documentElement.childNodes[1].childNodes[0].nodeValue
    except Exception:
        pass

    return code, message


def _exception_type_from_code(code: Optional[str]) -> Type:
    if code in (
        "ResourceGroupNotFound",
        "ResourceNotFound",
        "SecretNotFound",
        "NAME_UNKNOWN",
    ):
        return ResourceNotFoundError
    elif code in ("RoleAssignmentExists", "EntityAlreadyExists"):
        return ResourceExistsError
    elif code == "UpdateConditionNotSatisfied":
        return ResourceModifiedError
    else:
        return AzureRestApiError


async def raise_for_status(
    response: aiohttp.ClientResponse,
    ignored_status_codes: Iterable[Tuple[int, str]] = tuple(),
) -> None:
    """
    Like response.raise_for_status, but raises AzureRestApiError based on parsing the
    response content
    """
    if not response.ok:
        code = None
        message = None
        response_text = None

        if (
            response.headers.get("Content-Type", "")
            .lower()
            .startswith("application/json")
        ):
            try:
                code, message = _get_code_and_message_from_json(await response.json())
            except Exception:
                pass
        elif response.headers.get("Content-Type", "").startswith("application/xml"):
            try:
                response_text = await response.text()
            except Exception:
                response_text = ""

            code, message = _get_code_and_message_from_xml(response_text)

        if message is None:
            if response_text is None:
                # None means we haven't tried to get response.text. "" means we have
                # tried and the response was empty or we failed to get the text
                response_text = await response.text()
            message = response_text

        if code is None or (response.status, code) not in ignored_status_codes:
            raise _exception_type_from_code(code)(response.status, code, message)


def raise_for_status_sync(response: requests.Response) -> None:
    if not response.ok:
        code = None
        message = None
        response_text = None

        if (
            response.headers.get("Content-Type", "")
            .lower()
            .startswith("application/json")
        ):
            try:
                code, message = _get_code_and_message_from_json(response.json())
            except Exception:
                pass
        elif response.headers.get("Content-Type", "").startswith("application/xml"):
            try:
                response_text = response.text
            except Exception:
                response_text = ""

            code, message = _get_code_and_message_from_xml(response_text)

        if message is None:
            if response_text is None:
                response_text = response.text
            message = response_text

        raise _exception_type_from_code(code)(response.status_code, code, message)



"""
This code gets "embedded" into the user's code, so that the meadowrun server can invoke
a function in the user's code base.

This code should therefore have as few dependencies as possible (ideally none).

Lots of parallels to grid_worker.py
"""

import importlib  # available in python 3.1+
import argparse  # available in python 3.2+
import pickle
import traceback


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("--module-name")
    parser.add_argument("--function-name")
    parser.add_argument("--io-path", required=True)
    parser.add_argument("--has-pickled-function", action="store_true")
    parser.add_argument("--has-pickled-arguments", action="store_true")
    parser.add_argument("--result-highest-pickle-protocol", type=int, required=True)

    args = parser.parse_args()

    if bool(args.module_name) ^ bool(args.function_name):
        raise ValueError(
            "Cannot specify just one of --module-name and --function-name without the "
            "other"
        )
    if not (bool(args.module_name) ^ args.has_pickled_function):
        raise ValueError(
            "Must specify either --module-name with --function-name OR "
            "--has-pickled-function but not both"
        )

    # prepare filenames and pickle protocol for the result

    state_filename = args.io_path + ".state"
    result_filename = args.io_path + ".result"

    result_pickle_protocol = min(
        args.result_highest_pickle_protocol, pickle.HIGHEST_PROTOCOL
    )

    try:
        # import the module/unpickle the function

        if args.module_name is not None:
            print(f"About to import {args.module_name}.{args.function_name}")
            module = importlib.import_module(args.module_name)
            function = getattr(module, args.function_name)
            print(
                f"Imported {args.function_name} from "
                f"{getattr(module, '__file__', str(module))}"
            )
        else:
            pickled_function_path = args.io_path + ".function"
            print(f"Unpickling function from {pickled_function_path}")
            with open(pickled_function_path, "rb") as f:
                function = pickle.load(f)

        # read the arguments

        if args.has_pickled_arguments:
            with open(args.io_path + ".arguments", "rb") as f:
                # TODO probably should provide nicer error messages on unpickling
                function_args, function_kwargs = pickle.load(f)
        else:
            function_args, function_kwargs = (), {}

        # run the function
        result = function(*(function_args or ()), **(function_kwargs or {}))
        with open(result_filename, "wb") as f:
            pickle.dump(result, f, protocol=result_pickle_protocol)
    except Exception as e:
        # first print the exception for the local log file
        traceback.print_exc()

        # next, send the exception back
        tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
        # see MeadowRunClientAsync for why we don't just pickle the exception
        with open(state_filename, "w", encoding="utf-8") as state_text_writer:
            state_text_writer.write("PYTHON_EXCEPTION")
        with open(result_filename, "wb") as f:
            pickle.dump((str(type(e)), str(e), tb), f, protocol=result_pickle_protocol)
    else:
        # send back results
        with open(state_filename, "w", encoding="utf-8") as state_text_writer:
            state_text_writer.write("SUCCEEDED")
        # result already pickled above in the try block


if __name__ == "__main__":
    main()



from __future__ import annotations

import dataclasses
import datetime
import io
import json
import logging
from typing import List, Iterable, Dict, Tuple

import aiohttp
import boto3
import pkg_resources

from meadowrun.aws_integration.aws_core import _boto3_paginate
from meadowrun.config import (
    AVX,
    AvxVersion,
    GPU,
    GPU_MEMORY,
    INTEL_DEEP_LEARNING_BOOST,
    EVICTION_RATE_INVERSE,
    LOGICAL_CPU,
    MEMORY_GB,
)
from meadowrun.instance_selection import CloudInstanceType, ResourcesInternal
from meadowrun.local_cache import get_cached_json, save_json_to_cache, clear_cache


_CACHED_EC2_PRICES_FILENAME = "aws_ec2_prices.json"


async def _get_ec2_instance_types(region_name: str) -> List[CloudInstanceType]:
    """
    Gets a list of EC2 instance types and their prices in the format expected by
    agent_creator:choose_instance_types_for_job
    """

    # TODO at some point add cross-region optimization

    try:
        cached = get_cached_json(
            _CACHED_EC2_PRICES_FILENAME, datetime.timedelta(hours=4)
        )
        if cached:
            return [CloudInstanceType.from_dict(cit) for cit in cached]
    except Exception as e:
        print(
            "Warning, cached EC2 prices file appears to be created by an old version "
            f"of Meadowrun or is corrupted, will regenerate: {e}"
        )

    result = list(_get_ec2_on_demand_prices(region_name))
    on_demand_instance_types = {
        instance_type.name: instance_type for instance_type in result
    }
    result.extend(await _get_ec2_spot_prices(region_name, on_demand_instance_types))

    save_json_to_cache(_CACHED_EC2_PRICES_FILENAME, [cit.to_dict() for cit in result])

    return result


def _get_region_description_for_pricing(region_code: str) -> str:
    """
    Mostly copy/pasted from
    https://stackoverflow.com/questions/51673667/use-boto3-to-get-current-price-for-given-ec2-instance-type

    Converts something like us-east-2 to US East (Ohio). Almost all APIs use
    region_code, but the pricing API weirdly uses description.
    """
    endpoint_file = pkg_resources.resource_filename("botocore", "data/endpoints.json")
    with open(endpoint_file, "r", encoding="utf-8") as f:
        data = json.load(f)
    # Botocore is using Europe while Pricing API using EU...sigh...
    return data["partitions"][0]["regions"][region_code]["description"].replace(
        "Europe", "EU"
    )


def _get_ec2_on_demand_prices(region_name: str) -> Iterable[CloudInstanceType]:
    """
    Returns a dataframe with columns instance_type, memory_gb, logical_cpu, and price
    where price is the on-demand price
    """

    try:
        s3_client = boto3.client("s3", region_name="us-east-2")
        with io.BytesIO() as buffer:
            s3_client.download_fileobj("meadowrunprod", "aws_gpu_memory.json", buffer)
            buffer.seek(0)
            gpu_memory = json.loads(buffer.getvalue().decode("utf-8"))
    except Exception as e:
        logging.warning(
            "Unable to get most recent GPU memory data, will fall back to data in "
            f"package: {e}"
        )
        with open(
            pkg_resources.resource_filename(
                "meadowrun", "aws_integration/aws_gpu_memory.json"
            ),
            "r",
            encoding="utf-8",
        ) as file:
            gpu_memory = json.loads(file.read())

    # All comments about the pricing API are based on
    # https://www.sentiatechblog.com/using-the-ec2-price-list-api

    # us-east-1 is the only region this pricing API is available and the pricing
    # endpoint in us-east-1 has pricing data for all regions.
    pricing_client = boto3.client("pricing", region_name="us-east-1")

    filters = [
        # only get prices for the specified region
        {
            "Type": "TERM_MATCH",
            "Field": "location",
            "Value": _get_region_description_for_pricing(region_name),
        },
        # filter out instance types that come with SQL Server pre-installed
        {
            "Type": "TERM_MATCH",
            "Field": "preInstalledSw",
            "Value": "NA",
        },
        # limit ourselves to just Linux instances for now
        # TODO add support for Windows eventually
        {
            "Type": "TERM_MATCH",
            "Field": "operatingSystem",
            "Value": "Linux",
        },
        # Shared is a "regular" EC2 instance, as opposed to Dedicated and Host
        {"Type": "TERM_MATCH", "Field": "tenancy", "Value": "Shared"},
        # This relates to EC2 capacity reservations. Used is correct for when we don't
        # have any reservations
        {"Type": "TERM_MATCH", "Field": "capacitystatus", "Value": "Used"},
    ]

    for product_json in _boto3_paginate(
        pricing_client.get_products,
        Filters=filters,
        ServiceCode="AmazonEC2",
        FormatVersion="aws_v1",
    ):
        product = json.loads(product_json)
        attributes = product["product"]["attributes"]
        instance_type = attributes["instanceType"]

        consumable_resources = {}
        non_consumable_resources = {EVICTION_RATE_INVERSE: 100.0}

        # We don't expect the "warnings" to get hit, we just don't want to get thrown
        # off if the data format changes unexpectedly or something like that.

        if "physicalProcessor" not in attributes:
            print(
                f"Warning, skipping {instance_type} because physicalProcessor is not "
                "specified"
            )
            continue

        physical_processor = attributes["physicalProcessor"].lower()
        if "intel" in physical_processor:
            processor_brand = "intel"
        elif "amd" in physical_processor:
            processor_brand = "amd_cpu"
        elif "aws graviton" in physical_processor:
            processor_brand = "graviton"
        else:
            processor_brand = "other"
        # effectively, this skips Graviton (ARM-based) processors
        # TODO eventually support Graviton processors.
        if processor_brand not in ("intel", "amd_cpu"):
            # only log if we see non-Graviton processors
            if processor_brand != "graviton":
                print(
                    "Skipping non-Intel/AMD processor "
                    f"{attributes['physicalProcessor']} in {instance_type}"
                )
            continue
        non_consumable_resources[processor_brand] = 1.0

        if "OnDemand" not in product["terms"]:
            print(
                f"Warning, skipping {instance_type} because there was no OnDemand terms"
            )
            continue
        on_demand = list(product["terms"]["OnDemand"].values())
        if len(on_demand) != 1:
            print(
                f"Warning, skipping {instance_type} because there was more than one "
                "OnDemand SKU"
            )
            continue

        price_dimensions = list(on_demand[0]["priceDimensions"].values())
        if len(price_dimensions) != 1:
            print(
                f"Warning, skipping {instance_type} because there was more than one "
                "priceDimensions"
            )
            continue
        pricing = price_dimensions[0]

        if pricing["unit"] != "Hrs":
            print(
                f"Warning, skipping {instance_type} because the pricing unit is not "
                f"Hrs: {pricing['unit']}"
            )
            continue
        if "USD" not in pricing["pricePerUnit"]:
            print(
                f"Warning, skipping {instance_type} because the pricing is not in USD"
            )
            continue
        usd_price = pricing["pricePerUnit"]["USD"]

        try:
            usd_price_float = float(usd_price)
        except ValueError:
            print(
                f"Warning, skipping {instance_type} because the price is not a float: "
                f"{usd_price}"
            )
            continue

        memory = attributes["memory"]
        if not memory.endswith(" GiB"):
            print(
                f"Warning, skipping {instance_type} because memory doesn't end in GiB: "
                f"{memory}"
            )
            continue
        try:
            consumable_resources[MEMORY_GB] = float(memory[: -len(" GiB")])
        except ValueError:
            print(
                f"Warning, skipping {instance_type} because memory isn't an float: "
                f"{memory}"
            )
            continue

        try:
            consumable_resources[LOGICAL_CPU] = float(attributes["vcpu"])
        except ValueError:
            print(
                f"Warning, skipping {instance_type} because vcpu isn't a number: "
                f"{attributes['vcpu']}"
            )
            continue

        if "gpu" in attributes:
            try:
                consumable_resources[GPU] = float(attributes["gpu"])
            except ValueError:
                print(
                    f"Warning, not including gpu resources for {instance_type} because "
                    f"gpu isn't a number: {attributes['gpu']}"
                )
            else:
                # If we have a GPU, we should label what kind of GPU we have. All EC2
                # instance types have nvidia GPUs except for the g4ad family
                if instance_type.lower().startswith("g4ad"):
                    non_consumable_resources["amd_gpu"] = 1.0
                elif instance_type.lower().startswith("p2"):
                    # TODO p2 instances have a K80 GPU which isn't supported by the
                    # latest nvidia driver. We should add support for installing an old
                    # version of nvidia/cuda for use with these instances. Setting the
                    # "nvidia-old" flag is a bit of a hack and really just functions to
                    # hide these instance types from Meadowrun's allocation logic.
                    non_consumable_resources["nvidia-old"] = 1.0
                else:
                    non_consumable_resources["nvidia"] = 1.0

                # also, we'll try to assign it GPU memory from our data
                if instance_type.lower() not in gpu_memory:
                    print(
                        f"Warning, skipping {instance_type.lower()} because we don't "
                        "know how much GPU memory it has"
                    )
                else:
                    consumable_resources[GPU_MEMORY] = gpu_memory[instance_type.lower()]

        avx_version = AvxVersion.NONE
        for feature in attributes.get("processorFeatures", "").split("; "):
            if feature in ("Intel AVX", "AVX"):
                avx_version = max(avx_version, AvxVersion.AVX)
            elif feature in ("Intel AVX2", "AVX2"):
                avx_version = max(avx_version, AvxVersion.AVX2)
            elif feature == "Intel AVX512":
                avx_version = max(avx_version, AvxVersion.AVX512)
            elif feature == "Intel Deep Learning Boost":
                non_consumable_resources[INTEL_DEEP_LEARNING_BOOST] = 1.0
            # ignore unknown features
        if avx_version != AvxVersion.NONE:
            non_consumable_resources[AVX] = float(avx_version)

        yield CloudInstanceType(
            instance_type,
            "on_demand",
            usd_price_float,
            ResourcesInternal(consumable_resources, non_consumable_resources),
        )


async def _get_ec2_spot_prices(
    region_name: str, on_demand_instance_types: Dict[str, CloudInstanceType]
) -> List[CloudInstanceType]:
    """Returns a list of CloudInstanceTypes for spot instances"""
    ec2_client = boto3.client("ec2", region_name=region_name)

    # There doesn't appear to be an API for "give me the latest spot price for each
    # instance type". Instead, there's an API to get the spot price history. We query
    # for the last hour, assuming that all the instances we care about will have prices
    # within that last hour (no way to know whether that's actually true or not).
    start_time = datetime.datetime.utcnow() - datetime.timedelta(hours=1)
    # For each instance type, maps to the price, and the timestamp for that price. We're
    # going to always keep just the latest price. If we have multiple prices at the same
    # timestamp, we'll just take the largest one (this could happen because we get
    # different prices for different availability zones, e.g. us-east-2b vs us-east-2c).
    # TODO eventually account for AvailabilityZone?
    spot_prices: Dict[str, Tuple[float, datetime.datetime]] = {}
    for spot_price_record in _boto3_paginate(
        ec2_client.describe_spot_price_history,
        ProductDescriptions=["Linux/UNIX"],
        StartTime=start_time,
        MaxResults=10000,
    ):
        instance_type = spot_price_record["InstanceType"]
        timestamp = spot_price_record["Timestamp"]
        price = float(spot_price_record["SpotPrice"])
        prev_value = spot_prices.get(instance_type)

        if (
            prev_value is None
            or prev_value[1] < timestamp
            or (prev_value[1] == timestamp and prev_value[0] < price)
        ):
            spot_prices[instance_type] = price, timestamp

    # get interruption probabilities
    interruption_probabilities = await _get_ec2_interruption_probabilities(region_name)

    # TODO we should consider warning if we get spot prices or interruption
    # probabilities where we don't have on_demand_prices or spot_prices respectively,
    # right now we just drop that data

    results = []
    for instance_type, (price, timestamp) in spot_prices.items():
        # drop rows where we don't have the corresponding on_demand instance type
        # information
        if instance_type in on_demand_instance_types:
            on_demand_instance_type = on_demand_instance_types[instance_type]
            spot_instance_type = dataclasses.replace(
                on_demand_instance_type,
                price=price,
                on_demand_or_spot="spot",
                resources=on_demand_instance_type.resources.copy(),
            )
            # if interruption_probability is missing, just default to 80%
            spot_instance_type.resources.non_consumable[
                EVICTION_RATE_INVERSE
            ] = 100 - interruption_probabilities.get(instance_type, 80)
            results.append(spot_instance_type)

    return results


async def _get_ec2_interruption_probabilities(region_name: str) -> Dict[str, float]:
    """
    Returns a dataframe with columns instance_type, interruption_probability.
    interruption_probability is a percent, so values range from 0 to 100
    """

    # this is the data that drives https://aws.amazon.com/ec2/spot/instance-advisor/
    # according to
    # https://blog.doit-intl.com/spotinfo-a-new-cli-for-aws-spot-a9748bbe338f
    async with aiohttp.request(
        "GET", "https://spot-bid-advisor.s3.amazonaws.com/spot-advisor-data.json"
    ) as response:
        response.raise_for_status()
        data = await response.json()

    # The data we get isn't documented, but appears straightforward and can be checked
    # against the Spot Instance Advisor webpage. Each instance type gets an "r" which
    # corresponds to a range of interruption probabilities. The ranges are defined in
    # data["ranges"]. Each range has a "human readable label" like 15-20% and a "max"
    # like 22 (even though 20 != 22). We take an average interruption probability based
    # on the range implied by the maxes.

    # Get the average interruption probability for each range
    range_maxes = {r["index"]: r["max"] for r in data["ranges"]}
    range_keys_sorted = list(sorted(range_maxes.keys()))
    if range_keys_sorted != list(range(max(range_maxes.keys()) + 1)):
        raise ValueError(
            "Unexpected: ranges are not indexed contiguously from 0: "
            + ", ".join(str(key) for key in range_maxes.keys())
        )

    range_averages = {}
    for key in range_keys_sorted:
        if key == 0:
            range_min = 0
        else:
            range_min = range_maxes[key - 1]
        range_averages[key] = (range_maxes[key] + range_min) / 2

    # Get the average interruption probability for Linux instance_types in the specified
    # region
    return {
        instance_type: range_averages[values["r"]]
        for instance_type, values in data["spot_advisor"][region_name]["Linux"].items()
    }


def clear_prices_cache() -> None:
    clear_cache(_CACHED_EC2_PRICES_FILENAME)



"""
Copied from https://github.com/talkiq/gcloud-aio See sub-packages for version
information
"""

try:
    import pkg_resources

    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    __path__ = pkgutil.extend_path(__path__, __name__)  # type: ignore[has-type]



import datetime
import os
from typing import Any, Dict, Iterable

import boto3

from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _MEADOWRUN_GENERATED_DOCKER_REPO,
    ignore_boto3_error_code,
)

# queues will automatically be deleted 3 days after being created
_QUEUE_DELETION_TIMEOUT = datetime.timedelta(days=3)

# ECR images will be automatically deleted 2 days after last being created/used
_ECR_DELETION_TIMEOUT = datetime.timedelta(days=2)


def _get_meadowrun_task_queue_urls(sqs_client: Any) -> Iterable[str]:
    """Gets all SQS queue URLs created for meadowrun tasks"""
    for page in sqs_client.get_paginator("list_queues").paginate(
        QueueNamePrefix="meadowrun-task"
    ):
        if "QueueUrls" in page:
            for queue_url in page["QueueUrls"]:
                yield queue_url


def delete_old_task_queues(region_name: str) -> None:
    # TODO this would be better if the job clients could somehow keep their queues alive
    # while they're still being used rather than always having a very long timeout. We
    # could maybe abuse a random attribute which would probably cause the
    # LastModifiedTimestamp to update.
    now = datetime.datetime.utcnow()
    client = boto3.client("sqs", region_name=region_name)
    for queue_url in _get_meadowrun_task_queue_urls(client):
        success, response = ignore_boto3_error_code(
            lambda: client.get_queue_attributes(
                QueueUrl=queue_url,
                AttributeNames=[
                    "ApproximateNumberOfMessages",
                    "ApproximateNumberOfMessagesNotVisible",
                    "CreatedTimestamp",
                    "LastModifiedTimestamp",
                ],
            ),
            "AWS.SimpleQueueService.NonExistentQueue",
        )
        if not success:
            continue
        assert response is not None  # just for mypy
        attributes = response["Attributes"]
        # we should only need to check LastModifiedTimestamp, but we include
        # CreatedTimestamp just in case
        last_modified = max(
            datetime.datetime.fromtimestamp(int(attributes["CreatedTimestamp"])),
            datetime.datetime.fromtimestamp(int(attributes["LastModifiedTimestamp"])),
        )
        if (
            attributes["ApproximateNumberOfMessagesNotVisible"] == "0"
            and now - last_modified > _QUEUE_DELETION_TIMEOUT
        ):
            print(f"Deleting queue {queue_url}, was last modified at {last_modified}")
            client.delete_queue(QueueUrl=queue_url)


def delete_all_task_queues(region_name: str) -> None:
    """WARNING this will cause running run_map jobs to fail"""
    client = boto3.client("sqs", region_name=region_name)
    for queue_url in _get_meadowrun_task_queue_urls(client):
        ignore_boto3_error_code(
            lambda: client.delete_queue(QueueUrl=queue_url),
            "AWS.SimpleQueueService.NonExistentQueue",
        )


def delete_unused_images(region_name: str) -> None:
    """
    Deletes images in the _MEADOWRUN_GENERATED_DOCKER_REPO that haven't been
    pushed/pulled for _ECR_DELETION_TIMEOUT. The last pull time from AWS can be up to 24
    hours off (see describe_images documentation), so _ECR_DELETION_TIMEOUT should
    usually be about 24 hours longer than you actually want it to be.
    """
    images_to_delete = []

    client = boto3.client("ecr", region_name=region_name)

    success, result = ignore_boto3_error_code(
        lambda: client.describe_repositories(
            repositoryNames=[_MEADOWRUN_GENERATED_DOCKER_REPO]
        ),
        "RepositoryNotFoundException",
    )
    if not success:
        print(
            f"Repository {_MEADOWRUN_GENERATED_DOCKER_REPO} has not been created yet, "
            "so no images to clean up."
        )
        return

    cutoff_datetime = (
        datetime.datetime.now(datetime.timezone.utc) - _ECR_DELETION_TIMEOUT
    )
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr.html#ECR.Client.describe_images
    for page in client.get_paginator("describe_images").paginate(
        repositoryName=_MEADOWRUN_GENERATED_DOCKER_REPO
    ):
        for image in page["imageDetails"]:
            if "lastRecordedPullTime" in image:
                last_touched = image["lastRecordedPullTime"]
            else:
                last_touched = image["imagePushedAt"]

            if last_touched < cutoff_datetime:
                if "imageTags" not in image or len(image["imageTags"]) == 0:
                    print(
                        "Cannot delete image with no tags, you must delete it manually."
                        f" Digest is {image['imageDigest']}"
                    )
                else:
                    print(
                        "Will delete image with tag(s) "
                        f"{';'.join(image['imageTags'])} and digest "
                        f"{image['imageDigest']} which was last pulled/pushed at "
                        f"{last_touched}"
                    )
                    images_to_delete.append(image["imageTags"][0])

    if images_to_delete:
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecr.html#ECR.Client.batch_delete_image
        response = client.batch_delete_image(
            repositoryName=_MEADOWRUN_GENERATED_DOCKER_REPO,
            imageIds=[{"imageTag": tag} for tag in images_to_delete],
        )
        if response.get("failures"):
            raise ValueError(f"Failures: {response.get('failures')}")


def lambda_handler(event: Any, context: Any) -> Dict[str, Any]:
    """The handler for AWS lambda"""
    region_name = os.environ["AWS_REGION"]

    exceptions = []
    try:
        delete_old_task_queues(region_name)
    except Exception as e1:
        exceptions.append(e1)

    try:
        delete_unused_images(region_name)
    except Exception as e2:
        exceptions.append(e2)

    if len(exceptions) >= 1:
        if len(exceptions) > 1:
            # if only python had a way to combine exceptions...
            print(f"Ignoring second exception: {exceptions[1]})")
        raise exceptions[0]

    return {"statusCode": 200, "body": ""}



import asyncio.subprocess
import dataclasses
import datetime
import json
import os
import subprocess
import traceback
from typing import Optional, Tuple, Any, Dict

import aiohttp
import requests

from .azure_exceptions import raise_for_status, raise_for_status_sync


IMDS_AUTHORITY = "http://169.254.169.254"
IMDS_TOKEN_PATH = "/metadata/identity/oauth2/token"


@dataclasses.dataclass
class CachedAzureVariables:
    subscription_id: Optional[str]
    tenant_id: Optional[str]


CACHED_AZURE_VARIABLES = CachedAzureVariables(None, None)


_DEFAULT_SCOPE = "https://management.azure.com/.default"


def _scope_to_resource(scope: str) -> str:
    """
    Based on
    https://github.com/Azure/azure-sdk-for-python/blob/83964018f39b7702659d208cd2640f5eea7400fc/sdk/identity/azure-identity/azure/identity/_internal/__init__.py#L97

    which says "Convert an AADv2 scope to an AADv1 resource"
    """
    if scope.endswith("/.default"):
        return scope[: -len("/.default")]
    else:
        return scope


# loosely based on
# https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/azure-identity/azure/identity/_credentials/default.py

_CACHED_CLI_TOKEN: Optional[Tuple[str, datetime.datetime]] = None
# 5 minutes makes sense here because get-access-token is guaranteed to always give a
# token that has at least 5 minutes left on it
_CACHED_CLI_TOKEN_CUTOFF = datetime.timedelta(minutes=5)
_MANAGED_IDENTITY_CREDENTIAL: Optional[str] = None


async def get_token(scope: Optional[str] = None) -> str:
    global _MANAGED_IDENTITY_CREDENTIAL, _CACHED_CLI_TOKEN

    if scope is None:
        scope = _DEFAULT_SCOPE

    try:
        if scope == _DEFAULT_SCOPE:
            # If the user is logged in on the command line, will populate
            # _CACHED_CLI_TOKEN, CachedAzureVariables

            # if we don't have a cached token or the token is about to expire:
            if (
                _CACHED_CLI_TOKEN is None
                or _CACHED_CLI_TOKEN[1] - datetime.datetime.now()
                < _CACHED_CLI_TOKEN_CUTOFF
            ):
                token, expires_on, subscription_id, tenant_id = await _get_cli_token(
                    _DEFAULT_SCOPE
                )

                # according to https://github.com/Azure/azure-sdk-for-net/issues/15801
                # this will be a local datetime
                _CACHED_CLI_TOKEN = token, expires_on
                # not sure if this is the right behavior--need to figure out scenarios
                # where the subscription id would change
                if CACHED_AZURE_VARIABLES.subscription_id is None:
                    CACHED_AZURE_VARIABLES.subscription_id = subscription_id
                if CACHED_AZURE_VARIABLES.tenant_id is None:
                    CACHED_AZURE_VARIABLES.tenant_id = tenant_id

            return _CACHED_CLI_TOKEN[0]
        else:
            return (await _get_cli_token(scope))[0]
    except Exception:
        cli_exception = traceback.format_exc()

    try:
        if scope == _DEFAULT_SCOPE:
            if _MANAGED_IDENTITY_CREDENTIAL is None:
                _MANAGED_IDENTITY_CREDENTIAL = await _get_managed_identity_token(scope)

            return _MANAGED_IDENTITY_CREDENTIAL
        else:
            return await _get_managed_identity_token(scope)
    except Exception:
        managed_exception = traceback.format_exc()

    raise ValueError(
        "Unable to get a token for Azure. Error trying to get managed identity token:\n"
        f"{managed_exception}\n"
        f"Error trying to get CLI token:\n{cli_exception}"
    )


def get_token_sync(scope: Optional[str] = None) -> str:
    global _MANAGED_IDENTITY_CREDENTIAL, _CACHED_CLI_TOKEN

    if scope is None:
        scope = _DEFAULT_SCOPE

    try:
        if scope == _DEFAULT_SCOPE:
            if (
                _CACHED_CLI_TOKEN is None
                or _CACHED_CLI_TOKEN[1] - datetime.datetime.now()
                < _CACHED_CLI_TOKEN_CUTOFF
            ):
                token, expires_on, subscription_id, tenant_id = _get_cli_token_sync(
                    _DEFAULT_SCOPE
                )

                _CACHED_CLI_TOKEN = token, expires_on
                if CACHED_AZURE_VARIABLES.subscription_id is None:
                    CACHED_AZURE_VARIABLES.subscription_id = subscription_id
                if CACHED_AZURE_VARIABLES.tenant_id is None:
                    CACHED_AZURE_VARIABLES.tenant_id = tenant_id

            return _CACHED_CLI_TOKEN[0]
        else:
            return _get_cli_token_sync(scope)[0]
    except Exception:
        cli_exception = traceback.format_exc()

    try:
        if scope == _DEFAULT_SCOPE:
            if _MANAGED_IDENTITY_CREDENTIAL is None:
                _MANAGED_IDENTITY_CREDENTIAL = _get_managed_identity_token_sync(scope)

            return _MANAGED_IDENTITY_CREDENTIAL
        else:
            return _get_managed_identity_token_sync(scope)
    except Exception:
        managed_exception = traceback.format_exc()

    raise ValueError(
        "Unable to get a token for Azure. Error trying to get managed identity token:\n"
        f"{managed_exception}\n"
        f"Error trying to get CLI token:\n{cli_exception}"
    )


# roughly based on
# https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/azure-identity/azure/identity/_credentials/azure_cli.py

_CLI_TOKEN_EXPIRES_ON_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
_CLI_TOKEN_COMMAND_LINE = "az account get-access-token --output json --resource {}"


async def _get_cli_token(scope: str) -> Tuple[str, datetime.datetime, str, str]:
    """Returns token, expires on, subscription id, tenant id"""
    proc = await asyncio.subprocess.create_subprocess_shell(
        _CLI_TOKEN_COMMAND_LINE.format(_scope_to_resource(scope)),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )
    stdout, _ = await asyncio.wait_for(proc.communicate(), 10)
    output = stdout.decode()

    if proc.returncode != 0:
        raise ValueError(f"Unable to get CLI token: {output}")

    json_output = json.loads(output)

    return (
        json_output["accessToken"],
        datetime.datetime.strptime(
            json_output["expiresOn"], _CLI_TOKEN_EXPIRES_ON_FORMAT
        ),
        json_output["subscription"],
        json_output["tenant"],
    )


def _get_cli_token_sync(scope: str) -> Tuple[str, datetime.datetime, str, str]:
    result = subprocess.run(
        _CLI_TOKEN_COMMAND_LINE.format(_scope_to_resource(scope)),
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    output = result.stdout.decode()

    if result.returncode != 0:
        raise ValueError(f"Unable to get CLI token: {output}")

    json_output = json.loads(output)

    return (
        json_output["accessToken"],
        datetime.datetime.strptime(
            json_output["expiresOn"], _CLI_TOKEN_EXPIRES_ON_FORMAT
        ),
        json_output["subscription"],
        json_output["tenant"],
    )


# based on
# https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/azure-identity/azure/identity/_credentials/managed_identity.py


class EnvironmentVariables:
    AZURE_CLIENT_ID = "AZURE_CLIENT_ID"
    AZURE_TENANT_ID = "AZURE_TENANT_ID"

    AZURE_POD_IDENTITY_AUTHORITY_HOST = "AZURE_POD_IDENTITY_AUTHORITY_HOST"
    IDENTITY_ENDPOINT = "IDENTITY_ENDPOINT"
    IDENTITY_HEADER = "IDENTITY_HEADER"
    IDENTITY_SERVER_THUMBPRINT = "IDENTITY_SERVER_THUMBPRINT"
    IMDS_ENDPOINT = "IMDS_ENDPOINT"
    MSI_ENDPOINT = "MSI_ENDPOINT"
    MSI_SECRET = "MSI_SECRET"

    AZURE_AUTHORITY_HOST = "AZURE_AUTHORITY_HOST"

    AZURE_FEDERATED_TOKEN_FILE = "AZURE_FEDERATED_TOKEN_FILE"
    TOKEN_EXCHANGE_VARS = (
        AZURE_AUTHORITY_HOST,
        AZURE_TENANT_ID,
        AZURE_FEDERATED_TOKEN_FILE,
    )


def _get_managed_identity_token_helper(scope: str) -> Tuple[Dict[str, Any], bool]:
    scope = _scope_to_resource(scope)

    method = "GET"
    data = None
    disable_verify = False
    parameters = {}

    azure_client_id = os.environ.get(EnvironmentVariables.AZURE_CLIENT_ID)
    if azure_client_id is not None:
        parameters["client_id"] = azure_client_id

    if os.environ.get(EnvironmentVariables.IDENTITY_ENDPOINT):
        if os.environ.get(EnvironmentVariables.IDENTITY_HEADER):
            if os.environ.get(EnvironmentVariables.IDENTITY_SERVER_THUMBPRINT):
                # not tested: ServiceFabricCredential
                url = os.environ.get(EnvironmentVariables.IDENTITY_ENDPOINT)
                secret = os.environ.get(EnvironmentVariables.IDENTITY_HEADER)
                # thumbprint = os.environ.get(
                #     EnvironmentVariables.IDENTITY_SERVER_THUMBPRINT)
                # instead of using the thumbprint, it seems like we just ignore the
                # connection
                disable_verify = True
                headers = {"Secret": secret}
                parameters.update(
                    {"api-version": "2019-07-01-preview", "resource": scope}
                )
            else:
                # AppServiceCredential
                url = os.environ.get(EnvironmentVariables.IDENTITY_ENDPOINT)
                secret = os.environ.get(EnvironmentVariables.IDENTITY_HEADER)
                headers = {"X-IDENTITY-HEADER": secret}
                parameters.update({"api-version": "2019-08-01", "resource": scope})
        elif os.environ.get(EnvironmentVariables.IMDS_ENDPOINT):
            # not tested: AzureArcCredential
            raise NotImplementedError("AzureArcCredential is not implemented")
        else:
            # not tested: CloudShellCredential
            method = "POST"
            url = os.environ.get(EnvironmentVariables.MSI_ENDPOINT)
            headers = {"Metadata": "true"}
            data = {"resource": scope}
    elif os.environ.get(EnvironmentVariables.MSI_ENDPOINT):
        if os.environ.get(EnvironmentVariables.MSI_SECRET):
            # not tested: AzureMLCredential
            url = os.environ.get(EnvironmentVariables.MSI_ENDPOINT)
            secret = os.environ.get(EnvironmentVariables.MSI_SECRET)
            headers = {"secret": secret}
            parameters.update({"api-version": "2017-09-01", "resource": scope})
        else:
            # not tested: CloudShellCredential
            method = "POST"
            url = os.environ.get(EnvironmentVariables.MSI_ENDPOINT)
            headers = {"Metadata": "true"}
            data = {"resource": scope}
    elif all(os.environ.get(var) for var in EnvironmentVariables.TOKEN_EXCHANGE_VARS):
        raise NotImplementedError("Need to implement TokenExchangeCredential")
    else:
        url = (
            os.environ.get(
                EnvironmentVariables.AZURE_POD_IDENTITY_AUTHORITY_HOST, IMDS_AUTHORITY
            ).strip("/")
            + IMDS_TOKEN_PATH
        )
        parameters.update({"api-version": "2018-02-01", "resource": scope})
        headers = {"Metadata": "true"}

    return {
        "method": method,
        "url": url,
        "params": parameters,
        "headers": headers,
        "data": data,
    }, disable_verify


# on WSL, the IMDS endpoint will hang rather than failing to connect immediately, so put
# a relatively short timeout as it generally shouldn't require a long time to connect
_MANAGED_IDENTITY_TIMEOUT_SECS = 1


async def _get_managed_identity_token(scope: str) -> str:
    request_parameters, disable_verify = _get_managed_identity_token_helper(scope)
    if disable_verify:
        connector = aiohttp.TCPConnector(verify_ssl=False)
    else:
        connector = None

    async with aiohttp.request(
        **request_parameters,
        connector=connector,
        timeout=aiohttp.ClientTimeout(total=_MANAGED_IDENTITY_TIMEOUT_SECS),
    ) as response:
        await raise_for_status(response)
        return (await response.json())["access_token"]


def _get_managed_identity_token_sync(scope: str) -> str:
    request_parameters, disable_verify = _get_managed_identity_token_helper(scope)
    if disable_verify:
        request_parameters["verify"] = False

    response = requests.request(
        **request_parameters, timeout=_MANAGED_IDENTITY_TIMEOUT_SECS
    )
    raise_for_status_sync(response)
    return response.json()["access_token"]


async def get_current_user_id() -> str:
    """
    This functions gets the id of the current user that is signed in to the Azure CLI.

    In order to get this information, it looks like there are two different services,
    "Microsoft Graph" (developer.microsoft.com/graph) and "Azure AD Graph"
    (graph.windows.net), the latter being deprecated
    (https://devblogs.microsoft.com/microsoft365dev/microsoft-graph-or-azure-ad-graph/).
    I think these services correspond to two different python libraries, msal
    (https://docs.microsoft.com/en-us/python/api/overview/azure/active-directory?view=azure-python)
    and adal (https://docs.microsoft.com/en-us/python/api/adal/adal?view=azure-python),
    but these libraries don't appear to do anything super useful on their own.

    The deprecated Azure Graph API seems to correspond to a higher-level library
    azure-graphrbac, which does seem to have the functionality we need:
    azure.graphrbac.GraphRbacManagementClient.signed_in_user, but is deprecated along
    with Azure Graph
    (https://github.com/Azure/azure-sdk-for-python/issues/14022#issuecomment-752279618).

    The msgraph library that we use here seems to be a not-very-high-level library
    for Microsoft Graph (https://github.com/microsoftgraph/msgraph-sdk-python-core).

    As a side note, another way to get this information is to use the command line to
    call `az ad signed-in-user show`, but that appears to be relying on the deprecated
    Azure Graph API as it gives a deprecation warning.
    """

    token = await get_token("https://graph.microsoft.com")
    # https://docs.microsoft.com/en-us/graph/api/user-get?view=graph-rest-1.0&tabs=http
    async with aiohttp.request(
        "GET",
        "https://graph.microsoft.com/v1.0/me",
        headers={
            "Authorization": f"Bearer {token}",
        },
    ) as response:
        response.raise_for_status()
        return (await response.json())["id"]
